﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ursa.Extensions
{
    /// <summary>
    /// Framework extensions to the type byte[].
    /// </summary>
    internal static class ArrayByteExtension
    {
        public static ushort GetBits(this byte[] data, uint startBit, int totalBits)
        {

            if (totalBits < 0 || totalBits > sizeof(ushort)*8)
                throw new ArgumentException(String.Format("Bits to be retrieved must be between 0 and {0}.", sizeof(ushort)), "totalBits");
            ushort result = 0;
            if (totalBits == 0)
                return result;

            uint idx = startBit / 8;
            startBit %= 8;

            if (idx >= data.Length)
                throw new ArgumentOutOfRangeException(String.Format("Starting bit position {0} is out of range of the available bits (0-{1}).", startBit, (data.Length*8)-1), "startBit");

            for (int iBitCount = 0; iBitCount < totalBits && idx < data.Length; iBitCount++)
            {
                result <<= 1;
                if (((data[idx] << (byte) startBit) & 0x80) > 0)
                    result |= 0x01;
                startBit++;
                if (startBit == 8)
                {
                    startBit = 0;
                    idx++;
                }
            }
            return result;
        }
        
        public static void SetBits(this byte[] data, uint startBit, int totalBits, uint val)
        {
            int iIntBitSize = sizeof(uint) * 8;

            uint idx = startBit / 8;
            startBit %= 8;
            if (idx >= data.Length)
                throw new ArgumentOutOfRangeException(String.Format("Starting bit position {0} is out of range of the available bits (0-{1}).", startBit, (data.Length * 8) - 1), "startBit");

            val <<= (iIntBitSize - totalBits);


            for (int iBitCount = 0; iBitCount < totalBits && idx < data.Length; iBitCount++)
            {
                if ((val & (1 << iIntBitSize-1)) > 0)
                    data[idx] |= (byte) (1 << (int)(7 - startBit));
                else
                    data[idx] &= (byte) ~(1 << (int) (7 - startBit));

                val <<= 1;
                startBit++;
                if (startBit == 8)
                {
                    startBit = 0;
                    idx++;
                }
            }
        }

        public static string EncodeBitsIntoChars(this byte[] data, uint totalBits, uint bitsPerChar)
        {
            char[] result = new char[(int)Math.Ceiling((double)totalBits / (double)bitsPerChar) ];

            ;
            uint iByteCount = 0;
            int charIdx = 0;
            int bitShift = sizeof(ushort) * 8 - (int) bitsPerChar;

            for (uint iBitCount = 0; iBitCount < totalBits; charIdx++, iBitCount += bitsPerChar)
            {
                ushort binary = (ushort) ((data[iByteCount] << 8) | data[iByteCount + 1]);
                result[charIdx] = (char) (((binary >> bitShift) & 0x07) + '0');

                bitShift -= (int) bitsPerChar;
                if (bitShift < 0)
                {
                    bitShift += 8;
                    iByteCount++;
                }
            }

            return new string(result, 0, charIdx);
        }


        public static char[] ToCharArray(this byte[] data, int offset, int count)
        {
            if (offset+count > data.Length)
                throw new ArgumentException("The offset + count is greater than the length of the source array.");

            char[] dest = new char[count];
            for (int i = 0; i < count; i++)
                dest[i] = (char)(data[offset + i]);

            return dest;
        }

        public static string ToHexString(this byte[] data)
        {
            if (data.Length == 0)
                return "";

            return String.Join("", data.ToList().ConvertAll(b => String.Format("{0:x2}", b)));
        }

        public static void SetAll(this byte[] data, byte val)
        {
            data.SetAll(val, -1, -1);
        }

        public static void SetAll(this byte[] data, byte val, int start, int count)
        {
            if (start < 0)
                start = 0;

            if (count < 0)
                count = data.Length - start;

            if (start + count > data.Length)
                throw new ArgumentException("The value of start + count must be less than or equal to the length of data.");

            for (int i = 0; i < count; i++)
                data[i + start] = val;
        }

        public static void CopyTo(this byte[] src, int srcOffset, byte[] dest, int destOffset, int count)
        {
            if (dest == null)
                throw new ArgumentNullException("dest");

            if (srcOffset + count > src.Length)
                throw new ArgumentException("The srcOffset + count is greater than the length of the source array.");

            if (destOffset + count > dest.Length)
                throw new ArgumentException("The destOffset + count is greater than the length of the destination array.");

            for (int i = 0; i < count; i++)
                dest[destOffset + i] = src[srcOffset + i];
        }

        public static int CompareTo(this byte[] src, byte[] other, int count)
        {
            return src.CompareTo(0, other, 0, count);
        }

        public static int CompareTo(this byte[] src, int srcOffset, byte[] other, int otherOffset, int count)
        {
            if (other == null)
                return 1;

            if (srcOffset + count > src.Length)
                throw new ArgumentException("The srcOffset + count is greater than the length of the source array.");

            if (otherOffset + count > other.Length)
                throw new ArgumentException("The otherOffset + count is greater than the length of the other array.");

            for (int i = 0; i < count; i++)
            {
                if (src[srcOffset + i] > other[otherOffset + i])
                    return 1;
                else if (src[srcOffset + i] < other[otherOffset + i])
                    return -1;
            }

            return 0;
        }
    }
}
