﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using System.Security.Cryptography;
using System.IO;
using ursa.Extensions;

namespace ursa
{
    /// <summary>
    /// Represents a SecurID® token encoded in the SDTID format (XML)
    /// </summary>
    public class SecurIdSdtidToken : SecurIdToken
    {

        #region Fields

        /// <exclude />
        protected XmlDocument _xmlDoc;
        /// <exclude />
        protected XmlNode _xmlNodeHeader;
        /// <exclude />
        protected XmlNode _xmlNodeToken;
        /// <exclude />
        protected XmlNode _xmlNodeTrailer;
        /// <exclude />
        protected string _origin = "";
        /// <exclude />
        protected string _dest = "";
        /// <exclude />
        protected string _name = "";
        /// <exclude />
        protected byte[] _secret = null;
        /// <exclude />
        protected string _sSerialNoPadding = "";
        /// <exclude />
        protected byte[] _tokenMac;
        /// <exclude />
        protected byte[] _headerMac;
        /// <exclude />
        private static readonly int _iMaxHashData = 0x10000;

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdSdtidToken"/> class representing an empty token.
        /// </summary>
        public SecurIdSdtidToken() : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdSdtidToken"/> class with the value of an existing token.
        /// </summary>
        /// <param name="token">A <see cref="SecurIdToken"/> object representing a SecurID® token.  Ignored if <see langword="null"/>.</param>
        /// <overloads>Initializes a new instance of the <see cref="SecurIdSdtidToken"/> class.</overloads>
        public SecurIdSdtidToken(SecurIdToken token) : base(token)
        {
            _xmlDoc = null;
            _xmlNodeHeader = null;
            _xmlNodeToken = null;
            _xmlNodeTrailer = null;
            _origin = "";
            _dest = "";
            _name = "";
            _secret = null;
            _sSerialNoPadding = "";
            _tokenMac = null;
            _headerMac = null;
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Decrypts the token using the supplied password.
        /// </summary>
        /// <param name="pass">String containing the password for the token. If the token does not require a password, an empty string should be passed.</param>
        /// <param name="devid">This value is always ignored; SDTID tokens do not required a Device ID (Serial Number) for decryption.</param>
        /// <exception cref="ursaException">Thrown if the decryption fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override void Decrypt(string pass, string devid)
        {
            if (IsEmpty)
                throw new InvalidOperationException("Cannot decrypt; no valid token has been loaded.");

            ValidatePasswordForDecryption(pass);

            try
            {
                if (SerialNoPadding.Length == 0)
                    throw new ursaException(ursaError.XmlInvalid, "SN");

                if (Origin.Length == 0)
                    throw new ursaException(ursaError.XmlInvalid, "Origin");

                if (Dest.Length == 0)
                    throw new ursaException(ursaError.XmlInvalid, "Dest");

                if (Name.Length == 0)
                    throw new ursaException(ursaError.XmlInvalid, "Name");

                byte[] key0 = HashPassword(pass.Length > 0 ? pass : Origin, Dest, Name);
                byte[] key1 = DecryptSecret(Secret, Name, key0);

                byte[] batchMacHashKey = CalcuateKeyFromStrings("BatchMAC", Name, key1, BatchMacHashIV);
                byte[] tokenMacHashKey = CalcuateKeyFromStrings("TokenMAC", SerialNoPadding, key1, TokenMacHashIV);
                byte[] seedEncKey = CalcuateKeyFromStrings("TokenEncrypt", SerialNoPadding, key1, SeedEncryptionIV);

                byte[] mac0 = HashXml(_xmlNodeHeader, batchMacHashKey, BatchMacHashIV);
                byte[] mac1 = HashXml(_xmlNodeToken, tokenMacHashKey, TokenMacHashIV);

                bool mac0_passed = ursaEncryption.AesBlocksEqual(mac0, HeaderMAC);
                bool mac1_passed = ursaEncryption.AesBlocksEqual(mac1, TokenMAC);

                /* note that we cannot diagnose a corrupted <Secret> field */
                if (!mac0_passed && !mac1_passed)
                    if (pass.Length > 0)
                        throw new ursaException(ursaError.DecryptionFailed, "");
                    else
                        throw new ursaException(ursaError.PasswordRequired, "");

                if (!mac0_passed)
                    throw new ursaException(ursaError.DecryptionFailed, "Authentication of header using the header message authentication code (MAC) has failed.");

                if (!mac1_passed)
                    throw new ursaException(ursaError.DecryptionFailed, "Authentication of token using the token message authentication code (MAC) has failed.");

                _decryptedSeed = DecryptSeed(EncryptedSeed, SerialNoPadding, seedEncKey);
            }
            catch (ursaException ex)
            {
                // If decryption fails, prompt for a password
                if (ex.ErrorCode == ursaError.PasswordRequired)
                    _bFlagPasswordPrompt = true;
                throw ex;
            }

        }

        /// <summary>
        /// Creates a string representation of the token, which is encrypted using the supplied password.
        /// </summary>
        /// <param name="pass">String containing the new password for the token. If an empty string is supplied, the exported token will not require a password</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token, if any.</param>
        /// <returns>A string representation of the token in SDTID (XML) format.</returns>
        /// <exception cref="ursaException">Thrown if the export fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override string Export(string pass, string devid)
        {
            if (!HasDecryptedSeed)
                throw new ursaException(ursaError.TokenNeedsDecrypted);

            //byte[] secret = _secret;
            byte[] secret = ursaEncryption.GetRandomBytes(ursaEncryption.AesBlockSize);
            byte[] key0 = HashPassword(pass.Length > 0 ? pass : "N/A", "N/A", "N/A");
            byte[] key1 = DecryptSecret(secret, "N/A", key0);
            byte[] batchMacHashKey = CalcuateKeyFromStrings("BatchMAC", "N/A", key1, BatchMacHashIV);
            byte[] tokenMacHashKey = CalcuateKeyFromStrings("TokenMAC", Serial, key1, TokenMacHashIV);
            byte[] seedEncKey = CalcuateKeyFromStrings("TokenEncrypt", Serial, key1, SeedEncryptionIV);
            byte[] encryptSeed = DecryptSeed(DecryptedSeed, Serial, seedEncKey);

            XmlDocument doc = new XmlDocument();
            doc.AppendChild(doc.CreateXmlDeclaration("1.0", "", ""));

            XmlNode root = doc.AppendChild(doc.CreateElement("TKNBatch"));

            #region Token Header section

            XmlNode header = root.AppendChild(doc.CreateElement("TKNHeader"));

            header.AppendChild(doc.CreateElement("Version")).InnerText = "0";
            header.AppendChild(doc.CreateElement("Origin")).InnerText = "N/A";
            header.AppendChild(doc.CreateElement("Dest")).InnerText = "N/A";
            header.AppendChild(doc.CreateElement("Name")).InnerText = "N/A";
            header.AppendChild(doc.CreateElement("FirstToken")).InnerText = "N/A";
            header.AppendChild(doc.CreateElement("LastToken")).InnerText = "N/A";
            header.AppendChild(doc.CreateElement("NumTokens")).InnerText = "0";
            header.AppendChild(doc.CreateElement("Secret")).InnerText = Convert.ToBase64String(secret);
            header.AppendChild(doc.CreateElement("DefBirth")).InnerText = BirthDate.ToString("yyyy/MM/dd");
            header.AppendChild(doc.CreateElement("DefDeath")).InnerText = ExpirationDate.ToString("yyyy/MM/dd");
            header.AppendChild(doc.CreateElement("DefDigits")).InnerText = TokencodeLength.ToString();
            header.AppendChild(doc.CreateElement("DefInterval")).InnerText = IntervalSeconds.ToString();
            header.AppendChild(doc.CreateElement("DefAlg")).InnerText = Has128BitKey ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefMode")).InnerText = HasModeFlag ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefPrecision")).InnerText = "2400";
            header.AppendChild(doc.CreateElement("DefSmallWin")).InnerText = "630";
            header.AppendChild(doc.CreateElement("DefMediumWin")).InnerText = "4320";
            header.AppendChild(doc.CreateElement("DefLargeWin")).InnerText = "4320";
            header.AppendChild(doc.CreateElement("DefAddPIN")).InnerText = IsPinRequiredForTokencode ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefLocalPIN")).InnerText = IsLocalPin ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefCopyProtection")).InnerText = "1";
            header.AppendChild(doc.CreateElement("DefPinType")).InnerText = "0";
            header.AppendChild(doc.CreateElement("DefKeypad")).InnerText = "1";
            header.AppendChild(doc.CreateElement("DefProtLevel")).InnerText = "0";
            header.AppendChild(doc.CreateElement("DefRevision")).InnerText = "0";
            header.AppendChild(doc.CreateElement("DefTimeDerivedSeeds")).InnerText = UsesTimeDerivedSeeds ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefAppDerivedSeeds")).InnerText = UsesAppDerivedSeeds ? "1" : "0";
            header.AppendChild(doc.CreateElement("DefFormFactor")).InnerText = "20000001";

            byte[] headerMac = HashXml(header, batchMacHashKey, BatchMacHashIV);
            header.AppendChild(doc.CreateElement("HeaderMAC")).InnerText = Convert.ToBase64String(headerMac);

            #endregion


            #region Token section

            XmlNode tkn = root.AppendChild(doc.CreateElement("TKN"));

            tkn.AppendChild(doc.CreateElement("SN")).InnerText = Serial;

            //  the first character of <Seed> will be ignored by the reader
            tkn.AppendChild(doc.CreateElement("Seed")).InnerText = "=" + Convert.ToBase64String(encryptSeed);

            tkn.AppendChild(doc.CreateElement("UserFirstName")).InnerText = " ";
            tkn.AppendChild(doc.CreateElement("UserLastName")).InnerText = " ";
            tkn.AppendChild(doc.CreateElement("UserLogin")).InnerText = " ";

            XmlNode tokenAttributes = tkn.AppendChild(doc.CreateElement("TokenAttributes"));

            tokenAttributes.AppendChild(doc.CreateElement("DeviceSerialNumber")).InnerText = devid;
            tokenAttributes.AppendChild(doc.CreateElement("Nickname")).InnerText = " ";

            byte[] tokenMac = HashXml(tkn, tokenMacHashKey, TokenMacHashIV);
            tkn.AppendChild(doc.CreateElement("TokenMAC")).InnerText = Convert.ToBase64String(tokenMac);

            #endregion


            #region Token Trailer section

            byte[] byHashData = GetNodeContentsForHashing(true, header, tkn);
            byte[] batchSignature = ursaEncryption.RsaSha1SignData(BatchPrivateKey, byHashData);

            XmlNode trailer = root.AppendChild(doc.CreateElement("TKNTrailer"));

            trailer.AppendChild(doc.CreateElement("BatchSignature")).InnerText = Convert.ToBase64String(batchSignature);
            trailer.AppendChild(doc.CreateElement("BatchCertificate")).InnerText = Convert.ToBase64String(BatchCertificate);

            #endregion

            StringBuilder sbResult = new StringBuilder();
            using (XmlWriter xw = XmlWriter.Create(sbResult))
            {
                doc.WriteTo(xw);
            }
            return sbResult.ToString();
        }


        #endregion


        #region Protected Methods

        /// <summary>
        /// Decodes the SDTID (XML) string representation of a token.  This method is called by <see cref="SecurIdToken.Parse(string, int)"/> to initialize the current instance based on the values in the string representation.
        /// </summary>
        /// <param name="token">String containing a SDTID (XML) representation of a SecurID® token.</param>
        /// <param name="tokenNumber">Specifies which token will be parsed, if the SDTID format string contains more than one token.</param>
        /// <exception cref="ursaException">Thrown if the string can not be decoded.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        protected override void DecodeToken(string token, int tokenNumber)
        {
            bool strict = true;

            _xmlDoc = new XmlDocument();
            _xmlDoc.PreserveWhitespace = true;
            try
            {
                _xmlDoc.LoadXml(token);
            }
            catch (Exception ex)
            {
                throw new ursaException(ursaError.XmlInvalid, "XML parsing of sdtid failed.", ex);
            }

            XmlNode batch = _xmlDoc.SelectSingleNode("//TKNBatch");
            if (batch == null)
            {

                throw new ursaException(ursaError.XmlInvalid, "TKNBatch");
            }

            _xmlNodeHeader = batch.SelectSingleNode("TKNHeader");
            _xmlNodeTrailer = batch.SelectSingleNode("TKNTrailer");

            int idx = 0;
            foreach (XmlNode node in batch.ChildNodes)
            {
                if (node.NodeType == XmlNodeType.Element && node.Name == "TKN")
                {
                    if (tokenNumber == -1 && _xmlNodeToken != null)
                        throw new ursaException(ursaError.SdtidMultipleTokens);
                    if (idx++ == tokenNumber || tokenNumber == -1)
                        _xmlNodeToken = node;
                }
            }

            if (strict)
            {
                if (_xmlNodeHeader == null)
                    throw new ursaException(ursaError.XmlInvalid, "TKNHeader");
                if (_xmlNodeToken == null)
                    throw new ursaException(ursaError.XmlInvalid, "TKN");
                if (_xmlNodeTrailer == null)
                    throw new ursaException(ursaError.XmlInvalid, "TKNTrailer");
            }

            _sSerialNoPadding = FindXmlStringValue("SN");
            // hard token sdtid files are missing the leading zeros; pad as necessary
            _sSerial = String.Format(
                "{0}{1}",
                (_sSerialNoPadding.Length < SecurIdToken.SerialChars) ? new String('0', SecurIdToken.SerialChars - _sSerialNoPadding.Length) : "",
                _sSerialNoPadding);

            _bFlagAddPin = FindXmlIntValue("AddPIN", 0) > 0;
            _bFlagLocalPin = FindXmlIntValue("LocalPIN", 0) > 0;
            _iDigits = FindXmlIntValue("Digits", 6);

            try { _dtBirth = FindXmlDateValue("Birth"); }
            catch (Exception) { _dtExpiration = DateTime.MinValue; }

            try { _dtExpiration = FindXmlDateValue("Death"); }
            catch (Exception) { _dtExpiration = DateTime.MaxValue; }

            _encryptedSeed = FindXmlBase64Value("Seed");
            if (EncryptedSeed.Length != ursaEncryption.AesBlockSize)
                throw new ursaException(ursaError.XmlInvalid, "Seed");

            _bFlag128Bit = FindXmlIntValue("Alg", 0) > 0;
            _bFlagAppDerivedSeeds = FindXmlIntValue("AppDerivedSeeds", 0) > 0;
            _bFlagTimeDerivedSeeds = FindXmlIntValue("TimeDerivedSeeds", 0) > 0;
            _bFlagMode = FindXmlIntValue("Mode", 0) > 0;
            _bFlagSnPrompt = false;
            _iIntervalSeconds = FindXmlIntValue("Interval", 60);

            _origin = FindXmlStringValue("Origin");
            _dest = FindXmlStringValue("Dest");
            _name = FindXmlStringValue("Name");

            _tokenMac = FindXmlBase64Value("TokenMAC");
            if (_tokenMac.Length != ursaEncryption.AesBlockSize)
                throw new ursaException(ursaError.XmlInvalid, "TokenMAC");

            _headerMac = FindXmlBase64Value("HeaderMAC");
            if (_headerMac.Length != ursaEncryption.AesBlockSize)
                throw new ursaException(ursaError.XmlInvalid, "HeaderMAC");

            _secret = FindXmlBase64Value("Secret");
            if (_secret != null && _secret.Length > ursaEncryption.AesKeySize)
                throw new ursaException(ursaError.XmlInvalid, String.Format("Secret is longer than {0} bytes.", ursaEncryption.AesKeySize));

            _version = 0;
        }

        /// <summary>
        /// Retrieve the contents of all descendants of a given XML node, in a predefined format in preparation for hashing.
        /// </summary>
        /// <param name="parentNode">An <see cref="XmlNode"/> whose descendants will be retrieved.</param>
        /// <param name="includeMAC">If <see langword="true"/>, any MAC values in the XML will be included; if <see langword="false"/>, MAC values will be skipped.</param>
        /// <param name="nodeContents">A <see cref="StringBuilder"/> object to which the results will be appended.</param>
        /// <param name="padding">The number of padding characters to be appended to the results.</param>
        /// <param name="nodePrefix">The prefix to be prepended to each node name.</param>
        /// <returns>The number of child nodes found (direct children only).</returns>
        /// <remarks>
        /// This method is called by <see cref="GetNodeContentsForHashing"/> to build a string that will be subsequently hashed to create a MAC (message authentication code) or Signature.  This method calls itself recursively to retrieve all the descendant nodes.
        /// </remarks>
        protected int GetChildNodeContentsForHashing(XmlNode parentNode, bool includeMAC, StringBuilder nodeContents, ref int padding, string nodePrefix)
        {
            int children = 0;
            string longname = "";

            foreach (XmlNode node in parentNode.ChildNodes)
            {

                string val;
                string appendValue;
                int len;
                int ret;
                int remain = _iMaxHashData - nodeContents.Length;

                if (node.NodeType != XmlNodeType.Element)
                    continue;

                children++;

                if (!includeMAC && node.Name.EndsWith("MAC"))
                    continue;


                longname = String.Format("{0}.{1}", nodePrefix, node.Name);

                ret = GetChildNodeContentsForHashing(node, includeMAC, nodeContents, ref padding, longname);
                if (ret < 0)
                    return -1;
                if (ret > 0)
                    continue;

                val = node.InnerText;

                if (val.Length == 0)
                {
                    /*
        			 * An empty string is valid XML but it might violate
        			 * the sdtid format.  We'll handle it the same bizarre
        			 * way as RSA just to be safe.
        			 */
                    appendValue = String.Format("{0} </{1}>\n", longname, node.Name);
                    if (appendValue.Length > remain)
                        appendValue = appendValue.Substring(0, remain);

                    nodeContents.Append(appendValue);

                }
                else
                {
                    appendValue = String.Format("{0} {1}\n", longname, val);
                    if (appendValue.Length > remain)
                        appendValue = appendValue.Substring(0, remain);

                    nodeContents.Append(appendValue);

                    /* Bug compatibility :-( */
                    len = appendValue.Length + padding;
                    if (!includeMAC && len <= 16 && len < remain)
                    {
                        int insertPos = nodeContents.Length & ~0xf;
                        if (insertPos > 0)
                            nodeContents.Remove(insertPos, nodeContents.Length - insertPos);

                        appendValue = String.Format("{0} {1}\n", longname, node.Name);
                        if (appendValue.Length > remain)
                            appendValue = appendValue.Substring(0, remain);
                        nodeContents.Append(appendValue);
                    }
                }

                if (appendValue.Length >= remain)
                    return -1;

                /*
        		 * This doesn't really make sense but it's required for
        		 * compatibility
        		 */
                nodeContents.Append('\0', padding);
                if (!includeMAC)
                    padding = ((nodeContents.Length & 0xf) > 0) ? (nodeContents.Length & 0xf) : 0x10;
            }

            return children;
        }


        /// <summary>
        /// Retrieve the contents of the given XML nodes and all their descendants in a predefined format, and return the results as a byte array suitable for hashing.
        /// </summary>
        /// <param name="includeMAC">If <see langword="true"/>, any MAC values in the XML will be included; if <see langword="false"/>, MAC values will be skipped.</param>
        /// <param name="nodes">A list of <see cref="XmlNode"/> objects whose contents will be retrieved.</param>
        /// <returns>An array of bytes, containing the contents of the given XML nodes in a predefined format.</returns>
        /// <remarks>
        /// The return value of this method is subsequently hashed to create a MAC (message authentication code) or Signature. This method calls <see cref="GetChildNodeContentsForHashing"/> to recursively to retrieve the contents of all descendant nodes.
        /// </remarks>
        protected byte[] GetNodeContentsForHashing(bool includeMAC, params XmlNode[] nodes)
        {

            int padding = 0;
            StringBuilder sbHashString = new StringBuilder();

            foreach (XmlNode node in nodes)
                GetChildNodeContentsForHashing(node, includeMAC, sbHashString, ref padding, node.Name);

            if (sbHashString.Length > 0)
                return Encoding.ASCII.GetBytes(sbHashString.ToString());

            return new byte[0];
        }

        /// <summary>
        /// Calculate a hash value for an XML node and its descendants, using a supplied Key and IV (initialization vector)
        /// </summary>
        /// <param name="node">An <see cref="XmlNode"/> object to be hashed.</param>
        /// <param name="key">An AES key that will be used in the hash operation.</param>
        /// <param name="iv">An initialization vector (IV) that will be used in the hash operation.</param>
        /// <returns>An array of bytes containing the results of the hash.</returns>
        protected byte[] HashXml(XmlNode node, byte[] key, byte[] iv)
        {

            byte[] byHashData = GetNodeContentsForHashing(false, node);
            return ursaEncryption.AesCbcHash(key, iv, byHashData);
        }

        /// <summary>
        /// Calculate a hash value for a token password, using the supplied salt values.
        /// </summary>
        /// <param name="pass">A string containing a token password.</param>
        /// <param name="salt0">A string containing a salt value that will be used in the hash operation.</param>
        /// <param name="salt1">A string containing a second salt value that will be used in the hash operation.</param>
        /// <returns>An array of bytes containing the results of the hash.</returns>
        protected byte[] HashPassword(string pass, string salt0, string salt1)
        {

            byte[] result = new byte[ursaEncryption.AesBlockSize];
            byte[] key = new byte[ursaEncryption.AesKeySize];
            byte[] iv = new byte[ursaEncryption.AesBlockSize];
            byte[] tmp = new byte[ursaEncryption.AesBlockSize];
            byte[] data = new byte[0x50];

            key.SetAll(0);
            byte[] bySalt1 = Encoding.ASCII.GetBytes(salt1);
            bySalt1.CopyTo(0, key, 0, Math.Min(key.Length, bySalt1.Length));

            data.SetAll(0);
            byte[] bySalt0 = Encoding.ASCII.GetBytes(salt0);
            byte[] byPass = Encoding.ASCII.GetBytes(pass);
            byPass.CopyTo(0, data, 0, Math.Min(0x20, byPass.Length));
            bySalt0.CopyTo(0, data, 0x20, Math.Min(0x30, bySalt0.Length));

            result.SetAll(0);

            for (uint i = 0; i < 1000; i++)
            {
                data[0x4f] = (byte) (i >> 0);
                data[0x4e] = (byte) (i >> 8);

                tmp = ursaEncryption.AesCbcHash(key, iv, data);
                ursaEncryption.XorBlock(result, tmp);
            }

            return result;
        }

        /// <summary>
        /// Decrypt a token's Secret value.
        /// </summary>
        /// <param name="secret">An array of bytes containing a token Secret</param>
        /// <param name="partialIV">A string that will be used as part of the initialization vector (IV) for the decryption.</param>
        /// <param name="key">An array of bytes containing the decryption key.</param>
        /// <returns>An array of bytes containing the decrypted Secret.</returns>
        protected byte[] DecryptSecret(byte[] secret, string partialIV, byte[] key)
        {
            byte[] bySecret = Encoding.ASCII.GetBytes("Secret");
            byte[] byPartialIV = Encoding.ASCII.GetBytes(partialIV);

            byte[] iv = new byte[ursaEncryption.AesBlockSize];
            iv.SetAll(0);
            bySecret.CopyTo(0, iv, 0, bySecret.Length);
            byPartialIV.CopyTo(0, iv, 8, Math.Min(ursaEncryption.AesBlockSize - 8, byPartialIV.Length));

            byte[] result = ursaEncryption.AesECBEncrypt(key, iv);
            ursaEncryption.XorBlock(result, secret);
            return result;
        }

        /// <summary>
        /// Decrypt a token's Seed value.
        /// </summary>
        /// <param name="seed">An array of bytes containing a token's encrypted seed value</param>
        /// <param name="partialIV">A string that will be used as part of the initialization vector (IV) for the decryption.</param>
        /// <param name="key">An array of bytes containing the decryption key.</param>
        /// <returns>An array of bytes containing the decrypted seed.</returns>
        protected byte[] DecryptSeed(byte[] seed, string partialIV, byte[] key)
        {
            byte[] byStr0 = Encoding.ASCII.GetBytes(partialIV);
            byte[] bySecret = Encoding.ASCII.GetBytes("Seed");

            byte[] iv = new byte[ursaEncryption.AesBlockSize];
            iv.SetAll(0);
            byStr0.CopyTo(0, iv, 0, byStr0.Length);
            bySecret.CopyTo(0, iv, 8, Math.Min(ursaEncryption.AesBlockSize - 8, bySecret.Length));

            byte[] result = ursaEncryption.AesECBEncrypt(key, iv);
            ursaEncryption.XorBlock(result, seed);
            return result;
        }

        /// <summary>
        /// Calculate a key using two string values, an existing key, and an initialization vector (IV).
        /// </summary>
        /// <param name="string1">A string containing the first value that will be used to generate the key.</param>
        /// <param name="string2">A string containing the second value that will be used to generate the key.</param>
        /// <param name="key">An array of bytes containing an existing key.</param>
        /// <param name="iv">An array of bytes containing an initialization vector (IV).</param>
        /// <returns>An array of bytes containing the new key.</returns>
        protected byte[] CalcuateKeyFromStrings(string string1, string string2, byte[] key, byte[] iv)
        {

            byte[] buf = new byte[0x40];
            buf.SetAll(0);

            byte[] byString1 = Encoding.ASCII.GetBytes(string1);
            byte[] byString2 = Encoding.ASCII.GetBytes(string2);
            byString1.CopyTo(0, buf, 0, Math.Min(0x20, byString1.Length));
            byString2.CopyTo(0, buf, 0x20, Math.Min(0x20, byString2.Length));

            return ursaEncryption.AesCbcHash(key, iv, buf);
        }


        #region XML utility methods

        /// <summary>
        /// Search the XML of the current instance for a node with a given name and return the node's value, or an empty string if the node is not found.
        /// </summary>
        /// <param name="name">The name of the node to find.</param>
        /// <returns>The node's value, if found; an empty string otherwise.</returns>
        protected string FindXmlStringValue(string name)
        {
            return FindXmlStringValue(name, "");
        }

        /// <summary>
        /// Search the XML of the current instance for a node with a given name and return the node's value, or a default value if the node is not found.
        /// </summary>
        /// <param name="name">The name of the node to find.</param>
        /// <param name="defaultValue">The default value to return, if the node is not found.</param>
        /// <returns>The node's value, if found; the value of <paramref name="defaultValue"/> otherwise.</returns>
        /// <overloads>Search the XML of the current instance for a node with a given name and return the node's value.</overloads>
        protected string FindXmlStringValue(string name, string defaultValue)
        {

            XmlNode node = null;
            string value = "";

            if (_xmlNodeToken != null)
                node = _xmlNodeToken.SelectSingleNode("//" + name);

            if (node == null && _xmlNodeHeader != null)
                node = _xmlNodeHeader.SelectSingleNode(String.Format("//Def{0}", name));

            if (node == null && _xmlNodeHeader != null)
                node = _xmlNodeHeader.SelectSingleNode("//" + name);

            if (node != null)
                value = node.InnerText;

            if (value.Length == 0 && defaultValue.Length > 0)
                value = defaultValue;

            return value.Trim();
        }

        /// <summary>
        /// Search the XML of the current instance for a node with a given name and return the node's value as an <see cref="Int32"/> value, or a default value if the node is not found.
        /// </summary>
        /// <param name="name">The name of the node to find.</param>
        /// <param name="defaultValue">The default value to return, if the node is not found.</param>
        /// <returns>The node's value, if found; the value of <paramref name="defaultValue"/> otherwise.</returns>
        /// <exception cref="ursaException">Thrown if the node is found but the value cannot be converted into an <see cref="Int32"/> value.</exception>
        protected int FindXmlIntValue(string name, int defaultValue)
        {

            string ret = FindXmlStringValue(name);
            if (ret.Length == 0)
                return defaultValue;

            int val;
            try
            {
                val = Int32.Parse(ret);
            }
            catch (Exception ex)
            {
                throw new ursaException(ursaError.XmlInvalid, String.Format("Could not convert \"{0}\" to int.", ret), ex);
            }

            return val;
        }

        /// <summary>
        /// Search the XML of the current instance for a node with a given name and containing a base-64 encoded value, and decode the value into an array of bytes.
        /// </summary>
        /// <param name="name">The name of the node to find.</param>
        /// <returns>The node's value decoded into an array of bytes.  If the node is not found, or the node's value is not a valid base-64 string, an empty array is returned.</returns>
        protected byte[] FindXmlBase64Value(string name)
        {

            string data = FindXmlStringValue(name);
            if (data.Length == 0)
                return new byte[0];

            // <Seed> has a bogus character at the start of the string
            if (name == "Seed")
                data = data.Substring(1);

            byte[] bytes = null;

            try { bytes = Convert.FromBase64String(data); }
            catch (Exception) { bytes = new byte[0]; }

            return bytes;
        }

        /// <summary>
        /// Search the XML of the current instance for a node with a given name and containing a date value, and return the value as a <see cref="DateTime"/> object.
        /// </summary>
        /// <param name="name">The name of the node to find.</param>
        /// <returns>The node's value as a <see cref="DateTime"/> object, if found; <see name="DateTime.MinValue"/> otherwise.</returns>
        /// <exception cref="ursaException">Thrown if the node is found but the value cannot be converted into an <see cref="DateTime"/> object.</exception>
        protected DateTime FindXmlDateValue(string name)
        {

            string data = FindXmlStringValue(name);
            if (data.Length == 0)
                return DateTime.MinValue;

            DateTime dt;
            try
            {
                dt = DateTime.Parse(data);
            }
            catch (Exception ex)
            {
                throw new ursaException(ursaError.XmlInvalid, String.Format("Could not convert \"{0}\" to DateTime.", data), ex);
            }

            return dt;
        }

        #endregion


        #endregion


        #region Protected Properties

        /// <summary>
        /// Gets the Device ID / Serial Number from the SDTID token, without any added padding.
        /// </summary>
        protected string SerialNoPadding
        {
            get
            {
                return _sSerialNoPadding;
            }
        }

        /// <summary>
        /// Gets the message authentication code (MAC) for the SDTID token header
        /// </summary>
        protected byte[] HeaderMAC
        {
            get
            {
                return _headerMac;
            }
        }

        /// <summary>
        /// Gets the message authentication code (MAC) for the SDTID token body
        /// </summary>
        protected byte[] TokenMAC
        {
            get
            {
                return _tokenMac;
            }
        }

        /// <summary>
        /// Gets the Origin value of the SDTID token
        /// </summary>
        protected string Origin
        {
            get
            {
                return _origin;
            }
        }

        /// <summary>
        /// Gets the Dest value of the SDTID token
        /// </summary>
        protected string Dest
        {
            get
            {
                return _dest;
            }
        }

        /// <summary>
        /// Gets the Name value of the SDTID token
        /// </summary>
        protected string Name
        {
            get
            {
                return _name;
            }
        }

        /// <summary>
        /// Gets the Secret value of the SDTID token
        /// </summary>
        protected byte[] Secret
        {
            get
            {
                return _secret;
            }
        }

        /// <summary>
        /// Gets the initialization vector (IV) for calculating the Batch message authentication code (MAC) 
        /// </summary>
        protected byte[] BatchMacHashIV
        {
            get
            {
                return new byte[] { 0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c };
            }
        }

        //protected byte[] BatchEncHashIV
        //{
        //    get
        //    {
        //        return new byte[] { 0x32, 0x43, 0xf6, 0xa8, 0x88, 0x5a, 0x30, 0x8d, 0x31, 0x31, 0x98, 0xa2, 0xe0, 0x37, 0x07, 0x34 };
        //    }
        //}

        /// <summary>
        /// Gets the initialization vector (IV) for calculating the Token message authentication code (MAC) 
        /// </summary>
        protected byte[] TokenMacHashIV
        {
            get
            {
                return new byte[] { 0x1b, 0xb6, 0x7a, 0xe8, 0x58, 0x4c, 0xaa, 0x73, 0xb2, 0x57, 0x42, 0xd7, 0x07, 0x8b, 0x83, 0xb8 };
            }
        }

        /// <summary>
        /// Gets the initialization vector (IV) for Seed encryption 
        /// </summary>
        protected byte[] SeedEncryptionIV
        {
            get
            {
                return new byte[] { 0x16, 0xa0, 0x9e, 0x66, 0x7f, 0x3b, 0xcc, 0x90, 0x8b, 0x2f, 0xb1, 0x36, 0x6e, 0xa9, 0x57, 0xd3 };
            }
        }


        /// <summary>
        /// Gets the private key used to calculate the SDTID batch signature
        /// </summary>
        protected RSAParameters BatchPrivateKey
        {
            get
            {
                RSAParameters p = new RSAParameters();
                p.Modulus = new byte[] { 0xdd, 0xa1, 0xee, 0xa6, 0xd7, 0x66, 0xbb, 0xeb, 0xe2, 0x96, 0x0b, 0xeb, 0x19, 0x75, 0x44, 0x2b, 0x97, 0x1f, 0x25, 0x66, 0xb5, 0xc6, 0x03, 0x8d, 0x0f, 0xbb, 0x86, 0x91, 0xbb, 0x40, 0x04, 0xf9, 0x36, 0x4d, 0x04, 0xe7, 0x72, 0x4e, 0xca, 0x59, 0x84, 0xda, 0x2c, 0x64, 0xf2, 0xe5, 0x5f, 0xb4, 0x47, 0xc6, 0xf1, 0xe1, 0x53, 0xa5, 0xea, 0x15, 0x6d, 0xb4, 0x58, 0x51, 0xa4, 0xdd, 0x46, 0xc1, 0x22, 0x5f, 0xa1, 0x5c, 0xa5, 0xfb, 0x83, 0x9e, 0x72, 0x2b, 0xc0, 0xd4, 0x46, 0x69, 0x8e, 0x01, 0x35, 0x2c, 0x3f, 0x82, 0x57, 0x42, 0xf1, 0x38, 0x50, 0xc3, 0xf0, 0x6c, 0x1e, 0x28, 0xd1, 0x11, 0xe1, 0x32, 0xb5, 0x2c, 0xd0, 0x57, 0x06, 0x33, 0x54, 0xa3, 0x3a, 0x8e, 0x48, 0x26, 0xfa, 0x0b, 0xf1, 0x85, 0x52, 0xe8, 0xe8, 0x3e, 0x07, 0x6f, 0x54, 0x79, 0x68, 0xfe, 0x53, 0xc0, 0x01, 0x7b };
                p.Exponent = new byte[] { 0x01, 0x00, 0x01 };
                p.D = new byte[] { 0xca, 0x45, 0x93, 0xad, 0x29, 0x41, 0x55, 0x98, 0xbe, 0xbe, 0xfa, 0x39, 0xa2, 0x8e, 0x67, 0x9c, 0xf0, 0xdb, 0x38, 0x23, 0x39, 0x1a, 0x72, 0xfb, 0x36, 0xb4, 0x8a, 0xe8, 0x4d, 0xe0, 0xeb, 0xa9, 0x16, 0x69, 0xcc, 0x63, 0xfe, 0xea, 0xf1, 0xba, 0x29, 0x89, 0x84, 0xa2, 0xcd, 0x1b, 0x91, 0xf0, 0xd2, 0xe1, 0x3a, 0xb7, 0xce, 0xc3, 0xc9, 0x93, 0xac, 0xff, 0xbe, 0xeb, 0x24, 0x6f, 0xde, 0xb5, 0x8a, 0x37, 0xe2, 0x1f, 0xd4, 0x1e, 0x0b, 0x2e, 0xc4, 0xaf, 0x02, 0x73, 0xa7, 0xda, 0x33, 0x40, 0xa2, 0x22, 0xdc, 0x73, 0x63, 0x4b, 0xf2, 0xbd, 0xd0, 0x76, 0x18, 0xc8, 0xc5, 0xc6, 0x5a, 0xe7, 0x4a, 0xa9, 0x2f, 0xbb, 0xe1, 0xae, 0xe7, 0x3a, 0xcf, 0xd6, 0x4f, 0xa0, 0x58, 0x28, 0x3e, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
                p.P = new byte[] { 0xf7, 0x31, 0xfd, 0xe2, 0x7f, 0x4f, 0x9e, 0x3a, 0x61, 0x2e, 0x5d, 0x53, 0x1c, 0xc1, 0x9d, 0xc8, 0xa1, 0x69, 0xba, 0xe0, 0xc3, 0x01, 0x7e, 0x3d, 0xbe, 0xe0, 0x56, 0x81, 0x9c, 0x16, 0xe2, 0x53, 0x0f, 0xdc, 0xb1, 0xb6, 0xd1, 0x4a, 0xa2, 0x9c, 0x1c, 0x2e, 0x18, 0x53, 0x8d, 0x4e, 0x74, 0xe5, 0xb1, 0xd7, 0x69, 0xda, 0xf7, 0xff, 0xf5, 0xa3, 0x32, 0xd2, 0x37, 0x84, 0xe7, 0x86, 0xf0, 0x37 };
                p.Q = new byte[] { 0xe5, 0x86, 0xd9, 0xde, 0x9e, 0xdd, 0x7a, 0xbe, 0xb8, 0x0f, 0x2e, 0x2e, 0x34, 0x77, 0xd5, 0x89, 0x12, 0x29, 0x47, 0xbd, 0xe1, 0x4c, 0xc6, 0x28, 0xa9, 0x42, 0x38, 0x48, 0xa4, 0x47, 0xb8, 0xc2, 0x0b, 0xf8, 0x93, 0xe6, 0x81, 0x40, 0xe9, 0x04, 0xe1, 0x4f, 0x74, 0x46, 0x93, 0xfa, 0xeb, 0x40, 0x2f, 0x6c, 0x13, 0xf3, 0x70, 0x0f, 0xec, 0x3b, 0x59, 0xa4, 0xf6, 0x68, 0x4b, 0xcd, 0x6e, 0xdd };
                p.DP = new byte[] { 0x33, 0xe9, 0x70, 0xba, 0xd7, 0x27, 0x9e, 0x3f, 0xfe, 0x56, 0xa1, 0x4c, 0xa9, 0xf6, 0x53, 0x2f, 0x66, 0x0e, 0x71, 0x2b, 0x70, 0x68, 0x68, 0xdd, 0x88, 0xaf, 0x4e, 0x1b, 0x6b, 0xef, 0x36, 0x5a, 0x61, 0x33, 0x64, 0xb2, 0xd2, 0xe3, 0x0c, 0xa1, 0x22, 0x1d, 0xe0, 0x07, 0xf3, 0xdd, 0xed, 0x18, 0xab, 0xaf, 0x64, 0x50, 0x92, 0xd2, 0x53, 0x00, 0x91, 0xd4, 0xa9, 0xca, 0x24, 0x61, 0x27, 0x23 };
                p.DQ = new byte[] { 0x39, 0x3c, 0xba, 0xa2, 0x08, 0x6d, 0xe4, 0xc9, 0x20, 0xaf, 0x30, 0x6d, 0xf7, 0x49, 0x96, 0xe7, 0x7a, 0xae, 0xee, 0xa4, 0x0c, 0x46, 0x0f, 0xf8, 0x5d, 0xd5, 0x14, 0xa2, 0x10, 0xcd, 0x8d, 0xe6, 0x5e, 0x03, 0xdc, 0x26, 0x14, 0x3f, 0x72, 0x9c, 0x73, 0xef, 0x53, 0x68, 0xb2, 0x48, 0x55, 0x58, 0x09, 0x3b, 0x63, 0x72, 0x46, 0x94, 0xc1, 0xed, 0x3e, 0xfa, 0xa3, 0x33, 0xf9, 0x0b, 0x3e, 0xc5 };
                p.InverseQ = new byte[] { 0x89, 0x0b, 0xa9, 0x1d, 0xaf, 0x26, 0xc4, 0x50, 0xf4, 0xae, 0x69, 0x61, 0x85, 0xa5, 0x82, 0x62, 0x25, 0x81, 0x4f, 0x02, 0xc7, 0x9a, 0x0c, 0x64, 0x00, 0xc4, 0x8d, 0x5d, 0x40, 0x95, 0x5c, 0x99, 0x70, 0xb7, 0x5b, 0x79, 0x86, 0xd9, 0xeb, 0xa4, 0x72, 0x6d, 0xee, 0x01, 0xdc, 0x41, 0x88, 0x27, 0x21, 0x3f, 0x5a, 0x55, 0x9b, 0x1f, 0xd6, 0xe5, 0xdd, 0x38, 0x83, 0xb4, 0xa4, 0x22, 0xac, 0x8e };
                return p;
            }
        }

        /// <summary>
        /// Gets the SDTID batch certificate
        /// </summary>
        protected byte[] BatchCertificate
        {
            get
            {
                // You'd think this would contain an RSA modulus that matches BatchPrivateKey, but for whatever reason, it's completely different.
                return new byte[] { 0x30, 0x82, 0x02, 0x79, 0x30, 0x82, 0x01, 0x61, 0xa0, 0x03, 0x02, 0x01, 0x02, 0x02, 0x10, 0x33, 0x43, 0x45, 0x35, 0x35, 0x38, 0x35, 0x33, 0x33, 0x30, 0x39, 0x46, 0x38, 0x31, 0x33, 0x30, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x04, 0x05, 0x00, 0x30, 0x41, 0x31, 0x3f, 0x30, 0x3d, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x36, 0x53, 0x65, 0x63, 0x75, 0x72, 0x69, 0x74, 0x79, 0x20, 0x44, 0x79, 0x6e, 0x61, 0x6d, 0x69, 0x63, 0x73, 0x20, 0x54, 0x65, 0x63, 0x68, 0x6e, 0x6f, 0x6c, 0x6f, 0x67, 0x69, 0x65, 0x73, 0x2c, 0x20, 0x49, 0x6e, 0x63, 0x2e, 0x20, 0x50, 0x72, 0x69, 0x6d, 0x61, 0x72, 0x79, 0x20, 0x43, 0x41, 0x20, 0x52, 0x6f, 0x6f, 0x74, 0x20, 0x31, 0x30, 0x1e, 0x17, 0x0d, 0x30, 0x32, 0x30, 0x35, 0x31, 0x37, 0x31, 0x39, 0x32, 0x31, 0x35, 0x35, 0x5a, 0x17, 0x0d, 0x32, 0x32, 0x30, 0x35, 0x31, 0x32, 0x31, 0x39, 0x32, 0x31, 0x35, 0x35, 0x5a, 0x30, 0x34, 0x31, 0x32, 0x30, 0x30, 0x06, 0x03, 0x55, 0x04, 0x03, 0x13, 0x29, 0x53, 0x65, 0x63, 0x75, 0x72, 0x69, 0x74, 0x79, 0x20, 0x44, 0x79, 0x6e, 0x61, 0x6d, 0x69, 0x63, 0x73, 0x20, 0x54, 0x65, 0x63, 0x68, 0x6e, 0x6f, 0x6c, 0x6f, 0x67, 0x69, 0x65, 0x73, 0x20, 0x41, 0x43, 0x45, 0x2f, 0x53, 0x65, 0x72, 0x76, 0x65, 0x72, 0x30, 0x81, 0x9f, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x81, 0x8d, 0x00, 0x30, 0x81, 0x89, 0x02, 0x81, 0x81, 0x00, 0xd6, 0x7a, 0x75, 0x0c, 0x87, 0xf7, 0x1c, 0xe1, 0xc0, 0x2b, 0x66, 0xa1, 0x71, 0x1c, 0xd9, 0x08, 0x9b, 0x2a, 0x20, 0x2d, 0x50, 0x30, 0x4a, 0xad, 0xb1, 0xd6, 0xa7, 0x29, 0x21, 0x27, 0xe4, 0x21, 0xad, 0x2c, 0x27, 0x4b, 0xbf, 0xd2, 0xdb, 0x2d, 0x46, 0x28, 0xe9, 0xc4, 0x31, 0x29, 0x22, 0x6d, 0xc2, 0xf8, 0xa0, 0xa5, 0xe0, 0xe0, 0x04, 0x06, 0xff, 0x51, 0x87, 0x14, 0x35, 0x7c, 0xbf, 0xed, 0xd6, 0x3b, 0xac, 0x0e, 0x56, 0xa5, 0x89, 0x6c, 0x68, 0x0f, 0x61, 0xe4, 0x2f, 0x6a, 0xcc, 0xf3, 0x01, 0x1f, 0x15, 0x46, 0x48, 0x87, 0xb9, 0x93, 0xad, 0x6e, 0x51, 0xb0, 0x30, 0x74, 0x58, 0x42, 0x31, 0xf2, 0xe1, 0xef, 0xbe, 0xb0, 0x47, 0xff, 0xd5, 0x67, 0xb7, 0x2e, 0xca, 0x9a, 0x39, 0x77, 0x77, 0x02, 0x6e, 0xf8, 0x07, 0x1b, 0x0a, 0xb0, 0xe1, 0x9b, 0x84, 0x09, 0xb2, 0xaf, 0x38, 0xd3, 0x02, 0x03, 0x01, 0x00, 0x01, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x04, 0x05, 0x00, 0x03, 0x82, 0x01, 0x01, 0x00, 0xab, 0xc3, 0x0c, 0x26, 0xcd, 0x52, 0x73, 0x3c, 0x29, 0x7d, 0xc6, 0x6a, 0x9f, 0xd9, 0x68, 0x33, 0xed, 0x91, 0x84, 0x59, 0x8d, 0xd4, 0x86, 0x62, 0x5f, 0x27, 0xaf, 0xcf, 0x55, 0x6f, 0x18, 0x3f, 0x46, 0x52, 0xb9, 0x4b, 0x6a, 0x59, 0xbf, 0x12, 0xdf, 0x7c, 0xf7, 0x34, 0x7f, 0x75, 0x58, 0x5c, 0x88, 0x65, 0xe6, 0x00, 0x0c, 0x58, 0x9d, 0x5f, 0xcc, 0xa1, 0x99, 0xac, 0x86, 0x35, 0xe8, 0xce, 0x62, 0x60, 0x8d, 0x19, 0x9f, 0xaa, 0xa1, 0xcb, 0x4b, 0x8e, 0x94, 0x27, 0xd9, 0xcb, 0x67, 0x60, 0x6e, 0x11, 0x57, 0x2b, 0x1e, 0x79, 0xfa, 0x7f, 0xd9, 0x9e, 0x22, 0x37, 0x8c, 0x5a, 0x6f, 0x83, 0xbe, 0x1f, 0xbc, 0x56, 0x2f, 0x4a, 0x24, 0xe2, 0xe8, 0xcd, 0x86, 0xc3, 0xa1, 0xa5, 0x0c, 0x9b, 0xee, 0x45, 0xef, 0x2c, 0x0a, 0xb1, 0x4b, 0x81, 0x3d, 0x14, 0x89, 0xd7, 0x29, 0x06, 0x24, 0x6e, 0x9d, 0x38, 0x20, 0x59, 0x94, 0x91, 0xa6, 0x33, 0x79, 0x1b, 0x67, 0x9e, 0x25, 0x44, 0x02, 0xb6, 0x4d, 0x87, 0x2e, 0xd8, 0x93, 0x36, 0xd6, 0x2e, 0xc1, 0x29, 0xd7, 0x84, 0xf7, 0x12, 0x2d, 0xc9, 0xa6, 0xc9, 0xcc, 0x49, 0x37, 0x40, 0x2b, 0x17, 0xd1, 0xea, 0xd8, 0xee, 0x1d, 0xd5, 0xff, 0xfa, 0x24, 0x21, 0xf2, 0x96, 0x2e, 0x1f, 0x0b, 0x89, 0xea, 0x16, 0x6d, 0x5e, 0x3e, 0x56, 0xda, 0xd0, 0x12, 0x88, 0x94, 0xc3, 0x2a, 0x9d, 0x62, 0xcc, 0x30, 0x80, 0xc3, 0xb2, 0x46, 0x22, 0xae, 0x19, 0x39, 0x24, 0xdc, 0x38, 0x47, 0x76, 0x8a, 0xa1, 0x1b, 0xc4, 0xaa, 0x1d, 0x2c, 0x64, 0x3e, 0xda, 0x38, 0xda, 0x11, 0x27, 0xa4, 0xec, 0x7d, 0x8f, 0x6a, 0xea, 0x72, 0xc3, 0x96, 0xa2, 0xcb, 0xcd, 0xc9, 0xf9, 0xbd, 0x9f, 0x09, 0x2a, 0x25, 0xf8, 0x6e, 0x24, 0x29, 0x4b, 0xc0, 0xec, 0xf0, 0xe5 };
            }
        }


        #endregion


        #region Public Properties

        /// <summary>
        /// Returns <see langword="true"/> if this <see cref="SecurIdSdtidToken"/> object does not contain a valid token; <see langword="false"/> otherwise.
        /// </summary>
        public override bool IsEmpty
        {
            get
            {
                return base.IsEmpty || !HasEncryptedSeed;
            }
        }

        #endregion



    }

}
