﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ursa.Extensions;

namespace ursa
{
    /// <summary>
    /// Abstract base class for classes representing a SecurID® token
    /// </summary>
    public abstract class SecurIdToken
    {
        /// <summary>
        /// The expected length of a SecurID® Device ID / Serial Number
        /// </summary>
        protected const int SerialChars = 12;

        #region Fields

        /// <exclude />
        protected int _version;
        /// <exclude />
        protected string _sSerial;
        /// <exclude />
        protected bool _bIsSmartphone;
        /// <exclude />
        protected byte[] _encryptedSeed;
        /// <exclude />
        protected byte[] _decryptedSeed;
        /// <exclude />
        protected DateTime? _dtBirth;
        /// <exclude />
        protected DateTime? _dtExpiration;
        /// <exclude />
        protected int _iIntervalSeconds;
        /// <exclude />
        protected bool _bFlagLocalPin;
        /// <exclude />
        protected bool _bFlagAddPin;
        /// <exclude />
        protected int _iDigits;
        /// <exclude />
        protected bool _bFlagTimeDerivedSeeds;
        /// <exclude />
        protected bool _bFlagMode;
        /// <exclude />
        protected bool _bFlagAppDerivedSeeds;
        /// <exclude />
        protected bool _bFlagSnPrompt;
        /// <exclude />
        protected bool _bFlagPasswordPrompt;
        /// <exclude />
        protected bool _bFlag128Bit;
        /// <exclude />
        protected Regex _rgxNonAlphaNum;

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdToken"/> class with the value of an existing token.
        /// </summary>
        /// <param name="token">A <see cref="SecurIdToken"/> object representing a SecurID® token.  Ignored if <see langword="null"/>.</param>
        /// <remarks>This constructor is called by derived classes to preform common initialization.</remarks>
        protected SecurIdToken(SecurIdToken token)
        {
            _version = 0;
            _sSerial = "";
            _bIsSmartphone = false;
            _encryptedSeed = null;
            _decryptedSeed = null;
            _dtBirth = null;
            _dtExpiration = null;
            _iIntervalSeconds = 60;
            _bFlagLocalPin = false;
            _bFlagAddPin = false;
            _iDigits = 6;
            _bFlagTimeDerivedSeeds = false;
            _bFlagMode = false;
            _bFlagAppDerivedSeeds = false;
            _bFlagSnPrompt = false;
            _bFlagPasswordPrompt = false;
            _bFlag128Bit = false;

            _rgxNonAlphaNum = new Regex("[^0-9a-z]", RegexOptions.IgnoreCase | RegexOptions.Compiled);

            if (token != null)
                token.ShallowCopyTo(this);
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Converts the string representation of a SecurID® token to an instance of one of the classes derived from <see cref="SecurIdToken"/>.  If the string is in the sdtid format and contains more than one token, the first token is parsed.
        /// </summary>
        /// <param name="token">String containing a representation of a SecurID® token.  The string may be in any of the following formats: sdtid, version 2, or version 3.</param>
        /// <returns>An instance of one of the classes derived from <see cref="SecurIdToken"/>.  See remarks section for details.</returns>
        /// <exception cref="ursaException">Thrown if the string can not be parsed.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason why the parsing failed.</exception>
        /// <remarks>
        /// The value return by this method is an instance of one of the following classes derived from <see cref="SecurIdToken"/>:
        /// <list type="bullet">
        ///     <item>
        ///         <description><see cref="SecurIdSdtidToken"/></description>, if the string is in sdtid (XML) format
        ///     </item>
        ///     <item>
        ///         <description><see cref="SecurIdV2Token"/></description>, if the string is in version 2 format
        ///     </item>
        ///     <item>
        ///         <description><see cref="SecurIdV3Token"/></description>, if the string is in version 3 format
        ///     </item>
        /// </list>
        /// </remarks>
        /// <overloads>Converts the string representation of a SecurID® token to an instance of one of the classes derived from <see cref="SecurIdToken"/>.</overloads>
        public static SecurIdToken Parse(string token)
        {
            return Parse(token, -1);
        }

        /// <summary>
        /// Converts the string representation of a SecurID® token to an instance of one of the classes derived from <see cref="SecurIdToken"/>.  If the string is in the sdtid format and contains more than one token, the <paramref name="tokenNumber"/> parameter specifies which token will be parsed.
        /// </summary>
        /// <param name="token">String containing a representation of a SecurID® token.  The string may be in any of the following formats: sdtid, version 2, or version 3.</param>
        /// <param name="tokenNumber">Specifies which token will be parsed, if the string is in the sdtid format.  This value is ignored if the string is not in sdtid format.</param>
        /// <returns>An instance of one of the classes derived from <see cref="SecurIdToken"/>.  See remarks section for details.</returns>
        /// <exception cref="ursaException">Thrown if the string can not be parsed.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason why the parsing failed.</exception>
        /// <remarks>
        /// The value return by this method is an instance of one of the following classes derived from <see cref="SecurIdToken"/>:
        /// <list type="bullet">
        ///     <item>
        ///         <description><see cref="SecurIdSdtidToken"/></description>, if the string is in sdtid (XML) format
        ///     </item>
        ///     <item>
        ///         <description><see cref="SecurIdV2Token"/></description>, if the string is in version 2 format
        ///     </item>
        ///     <item>
        ///         <description><see cref="SecurIdV3Token"/></description>, if the string is in version 3 format
        ///     </item>
        /// </list>
        /// </remarks>
        public static SecurIdToken Parse(string token, int tokenNumber)
        {
            int iStartIndex = -1;
            bool bIsXml = false;


            string[] snippets = new string[] {
                "ctfData=3D", // broken quoted-printable input
                "ctfData=", // normal iPhone/Android soft token URLs
            };
            for (int i = 0; i < snippets.Length; i++)
            {
                int idx = token.IndexOf(snippets[i], StringComparison.CurrentCultureIgnoreCase);
                if (idx >= 0)
                {
                    iStartIndex = idx + snippets[i].Length;
                    break;
                }
            }

            if (iStartIndex < 0)
            {
                int idx = token.IndexOf("<?xml ", StringComparison.CurrentCultureIgnoreCase);
                if (idx >= 0)
                {
                    // sdtid (XML) token format
                    iStartIndex = idx;
                    bIsXml = true;
                }
            }

            if (iStartIndex < 0 && (token[0] == '1' || token[0] == '2' || token[0] == 'A'))
                iStartIndex = 0;

            if (iStartIndex < 0)
                throw new ursaException(ursaError.TokenStringInvalid);
            else if (iStartIndex > 0)
                token = token.Remove(0, iStartIndex);


            SecurIdToken t = null;
            if (bIsXml)
            {
                t = new SecurIdSdtidToken();
                t.DecodeToken(token, tokenNumber);
            }
            else if (token[0] == '1' || token[0] == '2')
            {
                t = new SecurIdV2Token();
                t.DecodeToken(token, 0);
            }
            else if (token[0] == 'A')
            {
                t = new SecurIdV3Token();
                t.DecodeToken(token, 0);
            }
            else
                throw new ursaException(ursaError.TokenStringInvalid);

            if (!bIsXml &&
                (token.StartsWith("com.rsa.securid.iphone://ctf", StringComparison.CurrentCultureIgnoreCase) ||
                token.StartsWith("com.rsa.securid://ctf", StringComparison.CurrentCultureIgnoreCase) ||
                token.StartsWith("http://127.0.0.1/securid/ctf", StringComparison.CurrentCultureIgnoreCase)))
                t._bIsSmartphone = true;

            return t;
        }

        /// <summary>
        /// Loads the string representation of a SecurID® token from a file, and calls <see cref="Parse(string, int)"/> to convert it to an instance of one of the classes derived from <see cref="SecurIdToken"/>.  If the file is in the sdtid format and contains more than one token, the first token is parsed.
        /// </summary>
        /// <param name="filename">The path of the file containing the SecurID® token.</param>
        /// <returns>An instance of one of the classes derived from <see cref="SecurIdToken"/>.  See <see cref="Parse(string, int)"/> for details.</returns>
        /// <exception cref="ursaException">Thrown if the file can not be read or if its contents cannot be parsed.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        /// <seealso cref="Parse(string, int)"/>
        public static SecurIdToken Load(string filename)
        {
            SecurIdToken token = null;
            try
            {
                FileStream fs = File.OpenRead(filename);

                if (fs != null)
                {
                    using (StreamReader rd = new StreamReader(fs))
                    {
                        string str = rd.ReadToEnd();
                        token = Parse(str, -1);
                    }
                }
            }
            catch (ursaException)
            {
                throw;
            }
            catch (Exception ex)
            {
                throw new ursaException(ursaError.FileReadFailed, "", ex);
            }
            return token;
        }

        /// <summary>
        /// Decrypts the token by passing the supplied password to the <see cref="Decrypt(string, string)"/> method.
        /// </summary>
        /// <param name="pass">String containing the password for the token.   If the token does not require a password, an empty string should be passed.</param>
        /// <exception cref="ursaException">Thrown if the decryption fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        /// <seealso cref="SecurIdSdtidToken.Decrypt(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV2Token.Decrypt(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV3Token.Decrypt(string, string)"  qualifyHint="true"/>
        public void Decrypt(string pass)
        {
            Decrypt(pass, "");
        }

        /// <summary>
        /// When implemented in a derived class, decrypts the token using the supplied password and Device ID (Serial Number).
        /// </summary>
        /// <param name="pass">String containing the password for the token. If the token does not require a password, an empty string should be passed.</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token.</param>
        /// <exception cref="ursaException">Thrown if the decryption fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        /// <seealso cref="SecurIdSdtidToken.Decrypt(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV2Token.Decrypt(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV3Token.Decrypt(string, string)"  qualifyHint="true"/>
        /// <remarks>
        /// This is an abstract method, implemented by each derived class.
        /// </remarks>
        /// <overloads>Decrypts the token using the supplied password and/or device ID.</overloads>
        public abstract void Decrypt(string pass, string devid);

        /// <summary>
        /// Computes the tokencode for a given date/time.
        /// </summary>
        /// <param name="time">The date/time for which the tokencode will be computed.</param>
        /// <returns>A string containing the tokencode for the supplied date/time.</returns>
        /// <exception cref="ursaException">Thrown if the tokencode cannot be computed.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public string ComputeTokencode(DateTime time)
        {
            return ComputeTokencode(time, "");
        }

        /// <summary>
        /// Computes the tokencode for a given date/time, using the specified PIN value.
        /// </summary>
        /// <param name="time">The date/time for which the tokencode will be computed.</param>
        /// <param name="pin">The PIN for the token, if required to compute the tokencode.</param>
        /// <returns>A string containing the tokencode for the supplied date/time.</returns>
        /// <exception cref="ursaException">Thrown if the tokencode cannot be computed.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        /// <overloads>Computes the tokencode for a given date/time.</overloads>
        public string ComputeTokencode(DateTime time, string pin)
        {
            if (IsPinRequiredForTokencode && String.IsNullOrWhiteSpace(pin))
                throw new ursaException(ursaError.PinRequired);

            if (!HasDecryptedSeed)
                throw new ursaException(ursaError.TokenNeedsDecrypted);

            byte[] key0;
            byte[] key1;

            /*
             * Convert the time to UTC and store as a packed binary-coded decimal (BCD)
             */
            DateTime timeUTC = time.ToUniversalTime();
            // If the interval is 60 seconds, drop the two least significant bits from the minute
            // If the interval is 30 seconds, drop the one least significant bit from the minute
            byte[] timeBCD = timeUTC.AddMinutes((timeUTC.Minute & (~(IntervalSeconds == 60 ? 0x03 : 0x01))) - timeUTC.Minute).ToPackedBCD(8);

            key0 = KeyFromBcdTime(timeBCD, 2);

            key0 = ursaEncryption.AesECBEncrypt(DecryptedSeed, key0);

            key1 = KeyFromBcdTime(timeBCD, 3);

            key1 = ursaEncryption.AesECBEncrypt(key0, key1);

            key0 = KeyFromBcdTime(timeBCD, 4);

            key0 = ursaEncryption.AesECBEncrypt(key1, key0);

            key1 = KeyFromBcdTime(timeBCD, 5);

            key1 = ursaEncryption.AesECBEncrypt(key0, key1);

            key0 = KeyFromBcdTime(timeBCD, 8);

            key0 = ursaEncryption.AesECBEncrypt(key1, key0);

            // key0 now contains 4 consecutive token codes; get the index for the correct code
            int idx = 0;
            if (IntervalSeconds == 60)
                idx = (timeUTC.Minute & 0x03) << 2;
            else
                idx = ((timeUTC.Minute & 0x01) << 3) | ((timeUTC.Second >= 30 ? 1 : 0) << 2);

            // get the correct token code
            uint tokencode = (uint) (key0[idx + 0] << 24) | (uint) (key0[idx + 1] << 16) | (uint) (key0[idx + 2] << 8) | (uint) (key0[idx + 3] << 0);

            string tokenString = tokencode.ToString();
            if (tokenString.Length > TokencodeLength)
                tokenString = tokenString.Substring(tokenString.Length - TokencodeLength);

            if (IsPinRequiredForTokencode && pin.Length > 0)
            {
                char[] caToken = tokenString.ToCharArray();
                for (int i = 0; i < pin.Length && i < caToken.Length; i++)
                {
                    caToken[caToken.Length - 1 - i] = (char)(((caToken[caToken.Length - 1 - i] - '0' + pin[pin.Length - 1 - i] - '0') % 10) + '0');
                }
                tokenString = new string(caToken);
            }

            return tokenString;
        }

        /// <summary>
        /// When implemented in a derived class, creates a string representation of the token, which is encrypted using the supplied password and Device ID (Serial Number)
        /// </summary>
        /// <param name="pass">String containing the new password for the token. If an empty string is supplied, the exported token will not require a password</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token, if any.</param>
        /// <returns>A string representation of the token.  The format of the string is determined by the implementation in the derived class.</returns>
        /// <exception cref="ursaException">Thrown if the export fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        /// <seealso cref="SecurIdSdtidToken.Export(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV2Token.Export(string, string)"  qualifyHint="true"/>
        /// <seealso cref="SecurIdV3Token.Export(string, string)"  qualifyHint="true"/>
        /// <remarks>
        /// This is an abstract method, implemented by each derived class.
        /// </remarks>
        public abstract string Export(string pass, string devid);

        /// <summary>
        /// Creates a string listing the values contained in this instance.
        /// </summary>
        /// <returns>A string listing the values contained in this instance.</returns>
        /// <remarks>
        /// The method creates a human-readable representation of the values contained in this token.  To export the token to a machine readable format, use <see cref="Export(string, string)"/>.
        /// </remarks>
        public override string ToString()
        {
            StringBuilder sbTokenInfo = new StringBuilder();
            sbTokenInfo.AppendFormat("Serial number:\t{0}{1}", Serial, Environment.NewLine);

            if (HasDecryptedSeed)
                sbTokenInfo.AppendFormat("Decrypted seed:\t{0}{1}", DecryptedSeed.ToHexString(), Environment.NewLine);

            if (HasEncryptedSeed)
                sbTokenInfo.AppendFormat("Encrypted seed:\t{0}{1}", EncryptedSeed.ToHexString(), Environment.NewLine);

            sbTokenInfo.AppendFormat("Password required:\t{0}{1}", IsPasswordRequired, Environment.NewLine);
            sbTokenInfo.AppendFormat("Device ID required:\t{0}{1}", IsDeviceIdRequired, Environment.NewLine);
            sbTokenInfo.AppendFormat("PIN required:\t{0}{1}", IsPinRequiredForTokencode, Environment.NewLine);
            sbTokenInfo.AppendFormat("Expiration date:\t{0}{1}", ExpirationDate, Environment.NewLine);
            sbTokenInfo.AppendFormat("Key length:\t{0}{1}", Has128BitKey ? "128" : "64", Environment.NewLine);
            sbTokenInfo.AppendFormat("Tokencode digits:\t{0}{1}", TokencodeLength, Environment.NewLine);
            sbTokenInfo.AppendFormat("Seconds per tokencode:\t{0}{1}", IntervalSeconds, Environment.NewLine);
            sbTokenInfo.AppendFormat("App-derived seeds:\t{0}{1}", UsesAppDerivedSeeds, Environment.NewLine);
            sbTokenInfo.AppendFormat("Time-derived seeds:\t{0}{1}", UsesTimeDerivedSeeds, Environment.NewLine);
            sbTokenInfo.AppendFormat("Local PIN:\t{0}{1}", IsLocalPin, Environment.NewLine);
            sbTokenInfo.AppendFormat("Mode:\t{0}{1}", HasModeFlag, Environment.NewLine);

            return sbTokenInfo.ToString();
        }

        #endregion


        #region Protected Methods

        /// <summary>
        /// When implemented in a derived class, decodes the string representation of a token.  This method is called by <see cref="Parse(string, int)"/> to initialize the current instance based on the values in the string representation.
        /// </summary>
        /// <param name="token">String containing a representation of a SecurID® token.  The required format is determined by the implementation of this method in each derived class.</param>
        /// <param name="tokenNumber">Specifies which token will be parsed. This applies only if the required format is sdtid, otherwise this parameter is ignored.</param>
        /// <remarks>
        /// This is an abstract method, implemented by each derived class.
        /// </remarks>
        protected abstract void DecodeToken(string token, int tokenNumber);

        /// <summary>
        /// Creates an encryption key value based on a specified number of bytes from a date/time value that is encoded as a packed binary-coded decimal (BCD)
        /// </summary>
        /// <param name="timeBCD">A date/time value encoded as a packed binary-coded decimal (BCD)</param>
        /// <param name="byteCount">The number of bytes from <paramref name="timeBCD"/> to be included in the key.</param>
        /// <returns>An array of bytes containing the new key.</returns>
        protected byte[] KeyFromBcdTime(byte[] timeBCD, int byteCount)
        {
            byte[] key = new byte[ursaEncryption.AesKeySize];

            // initialize bytes 0-7 of the key to 0xAA
            for (int i = 0; i < 8; i++)
                key[i] = 0xAA;

            // copy [byteCount] bytes of [timeBCD] to the key
            timeBCD.CopyTo(0, key, 0, byteCount);

            // convert the 5th-12th digits of the serial number to packed BCD, 
            // and copy to bytes 8-11 of the key
            int keyIdx = 8;
            for (int i = 4; i < 12; i += 2, keyIdx++)
                key[keyIdx] = (byte) (((Serial[i] - '0') << 4) | (Serial[i + 1] - '0'));

            // set bytes 12-15 of the key to 0xBB
            for (int i = 12; i < 16; i++)
                key[i] = 0xBB;

            return key;
        }

        /// <summary>
        /// Checks if the supplied Device ID is valid for use in decrypting this token, and removes any invalid characters from the Device ID.
        /// </summary>
        /// <param name="deviceId">A string containing a Device ID (Serial Number)</param>
        /// <returns>A copy of the Device ID will all invalid characters removed, if a Device ID is required for decryption.  Returns an empty string if a Device ID is not required for decryption.</returns>
        /// <exception cref="ursaException">Thrown with a <see cref="ursaException.ErrorCode"/> value of <see cref="ursaError.DeviceIdRequired" qualifyHint="true"/> if a Device ID is required for decryption and <paramref name="deviceId"/> is null, empty, or whitespace.</exception>
        protected string ValidateDeviceIDForDecryption(string deviceId)
        {
            if (IsDeviceIdRequired)
            {
                if (String.IsNullOrWhiteSpace(deviceId))
                    throw new ursaException(ursaError.DeviceIdRequired);
                return ScrubDeviceID(deviceId);
            }
            else
                return "";
        }

        /// <summary>
        /// Checks if the supplied password is valid for use in decrypting this token.
        /// </summary>
        /// <param name="password">A string containing a password.</param>
        /// <returns>A copy of the password, if a password is required for decryption.  Returns an empty string if a password is not required for decryption.</returns>
        /// <exception cref="ursaException">Thrown with a <see cref="ursaException.ErrorCode"/> value of <see cref="ursaError.PasswordRequired" qualifyHint="true"/> if a password is required for decryption and <paramref name="password"/> is null, empty, or whitespace.</exception>
        protected string ValidatePasswordForDecryption(string password)
        {
            if (IsPasswordRequired)
            {
                if (String.IsNullOrWhiteSpace(password))
                    throw new ursaException(ursaError.PasswordRequired);
                return password;
            }
            else
                return "";
        }

        /// <summary>
        /// Converts a SecurID® date value (unsigned short) to a DateTime instance.
        /// </summary>
        /// <param name="securdIdDate">An unsigned short value representing a date/time as the number of days since January 1, 2000 UTC.</param>
        /// <returns>A DateTime instance initialized to the date/time represented by <paramref name="securdIdDate"/>.</returns>
        /// <seealso cref="ConvertToSecuridDate"/>
        protected DateTime ConvertSecuridDate(ushort securdIdDate)
        {
            return new DateTime(2000, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddDays(securdIdDate);
        }

        /// <summary>
        /// Converts a DateTime instance to a SecurID® date value.
        /// </summary>
        /// <param name="date">The DateTime instance to be converted.</param>
        /// <returns>The number of days difference between <paramref name="date"/> and January 1, 2000 UTC.</returns>
        /// <seealso cref="ConvertSecuridDate"/>
        protected ushort ConvertToSecuridDate(DateTime date)
        {
            return (ushort)(date - new DateTime(2000, 1, 1, 0, 0, 0, DateTimeKind.Utc)).TotalDays;
        }

        /// <summary>
        /// Remove all invalid characters from a SecurID® Device ID (Serial Number)
        /// </summary>
        /// <param name="deviceId">A string containing a SecurID® Device ID (Serial Number)</param>
        /// <returns>A copy of <paramref name="deviceId"/> with all invalid characters removed.</returns>
        protected string ScrubDeviceID(string deviceId)
        {
            return _rgxNonAlphaNum.Replace(deviceId, "").ToUpper();
        }

        /// <summary>
        /// Perform a shallow copy of the current instance to supplied token instance
        /// </summary>
        /// <param name="token">The token instance into which the value of the current instance will be copied.</param>
        protected void ShallowCopyTo(SecurIdToken token)
        {
            if (token == null)
                throw new ArgumentNullException("token");

            token._version = _version;
            token._sSerial = _sSerial;
            token._bIsSmartphone = _bIsSmartphone;
            token._encryptedSeed = _encryptedSeed;
            token._decryptedSeed = _decryptedSeed;
            token._dtBirth = _dtBirth;
            token._dtExpiration = _dtExpiration;
            token._iIntervalSeconds = _iIntervalSeconds;
            token._bFlagLocalPin = _bFlagLocalPin;
            token._bFlagAddPin = _bFlagAddPin;
            token._iDigits = _iDigits;
            token._bFlagTimeDerivedSeeds = _bFlagTimeDerivedSeeds;
            token._bFlagMode = _bFlagMode;
            token._bFlagAppDerivedSeeds = _bFlagAppDerivedSeeds;
            token._bFlagSnPrompt = _bFlagSnPrompt;
            token._bFlagPasswordPrompt = _bFlagPasswordPrompt;
            token._bFlag128Bit = _bFlag128Bit;
        }

        #endregion


        //int securid_check_exp(struct securid_token *t, time_t now)
        //{
        //	time_t exp_unix_time = securid_unix_exp_date(t);
        //	const int halfday = 60 * 60 * 12, wholeday = 60 * 60 * 24;

        //	/*
        //	 * Other soft token implementations seem to allow ~12hrs as a grace
        //	 * period.  Actual results will depend on how soon the server cuts
        //	 * off expired tokens.
        //	 */
        //	exp_unix_time += halfday;
        //	exp_unix_time -= now;
        //	return exp_unix_time / wholeday;


        #region Properties

        /// <summary>
        /// Returns <see langword="true"/> if this <see cref="SecurIdToken"/> object does not contain a valid token; <see langword="false"/> otherwise.
        /// </summary>
        public virtual bool IsEmpty
        {
            get
            {
                return Serial.Length == 0;
            }
        }

        /// <summary>
        /// Gets the Device ID / Serial Number of the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public string Serial
        {
            get
            {
                return _sSerial;
            }
        }

        /// <summary>
        /// Gets the encrypted seed value of the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public byte[] EncryptedSeed
        {
            get
            {
                return _encryptedSeed;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if <see cref="EncryptedSeed"/> contains a seed value; <see langword="false"/> otherwise.
        /// </summary>
        public bool HasEncryptedSeed
        {
            get
            {
                return (EncryptedSeed != null && EncryptedSeed.Length > 0);
            }
        }

        /// <summary>
        /// Gets the decrypted seed value of the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public byte[] DecryptedSeed
        {
            get
            {
                return _decryptedSeed;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if <see cref="DecryptedSeed"/> contains a seed value; <see langword="false"/> otherwise.
        /// </summary>
        public bool HasDecryptedSeed
        {
            get
            {
                return (DecryptedSeed != null && DecryptedSeed.Length > 0);
            }
        }

        /// <summary>
        /// Gets the version number of the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public int Version
        {
            get
            {
                return _version;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token that uses app-derived seeds; <see langword="false"/> otherwise.
        /// </summary>
        public bool UsesAppDerivedSeeds
        {
            get
            {
                return _bFlagAppDerivedSeeds;
            }
        }


        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token that uses time-derived seeds; <see langword="false"/> otherwise.
        /// </summary>
        public bool UsesTimeDerivedSeeds
        {
            get
            {
                return _bFlagTimeDerivedSeeds;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token with the Mode flag set; <see langword="false"/> otherwise.
        /// </summary>
        public bool HasModeFlag
        {
            get
            {
                return _bFlagMode;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token with a 128-bit key; <see langword="false"/> if it has a 64-bit key.
        /// </summary>
        public bool Has128BitKey
        {
            get
            {
                return _bFlag128Bit;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if a Device ID (Serial Number) is required for the token represented by the current <see cref="SecurIdToken"/> object; <see langword="false"/> otherwise.
        /// </summary>
        public bool IsDeviceIdRequired
        {
            get
            {
                return _bFlagSnPrompt;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if a Password is required for the token represented by the current <see cref="SecurIdToken"/> object; <see langword="false"/> otherwise.
        /// </summary>
        public bool IsPasswordRequired
        {
            get
            {
                return _bFlagPasswordPrompt;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if a PIN is required to compute the tokencode for the current <see cref="SecurIdToken"/> object; <see langword="false"/> otherwise.
        /// </summary>
        public bool IsPinRequiredForTokencode
        {
            get
            {
                return _bFlagAddPin;
            }
        }


        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token intended for use on a smartphone; <see langword="false"/> otherwise.
        /// </summary>
        public bool IsSmartPhone
        {
            get
            {
                return _bIsSmartphone;
            }
        }

        /// <summary>
        /// Returns <see langword="true"/> if the current <see cref="SecurIdToken"/> object represents a token with a local PIN; <see langword="false"/> otherwise.
        /// </summary>
        public bool IsLocalPin
        {
            get
            {
                return _bFlagLocalPin;
            }
        }

        /// <summary>
        /// Gets the number of digits in each tokencode generated by the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public int TokencodeLength
        {
            get
            {
                return _iDigits;
            }
        }

        /// <summary>
        /// Gets the number seconds between distinct tokencodes generated by the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public int IntervalSeconds
        {
            get
            {
                return _iIntervalSeconds;
            }
        }

        /// <summary>
        /// Gets the expiration date of the token represented by the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public DateTime ExpirationDate
        {
            get
            {
                if (_dtExpiration == null)
                    return DateTime.MaxValue;
                return _dtExpiration.Value;
            }

        }

        /// <summary>
        /// Gets the birthdate of the token represented by the current <see cref="SecurIdToken"/> object.
        /// </summary>
        public DateTime BirthDate
        {
            get
            {
                if (_dtBirth == null)
                    return DateTime.MinValue;
                return _dtBirth.Value;
            }

        }

        /// <summary>
        /// Gets a <see cref="SecurIdV2Token"/> object representing the same token as the current <see cref="SecurIdToken"/> object.
        /// </summary>
        /// <remarks>
        /// If the current <see cref="SecurIdToken"/> object is an instance of the <see cref="SecurIdV2Token"/> class, then the current instance is returned.  Otherwise, a new instance of the <see cref="SecurIdV2Token"/> class is returned, initialized with the values from the current <see cref="SecurIdToken"/> object.
        /// </remarks>
        public SecurIdV2Token AsV2Token
        {
            get
            {
                if (this is SecurIdV2Token)
                    return (SecurIdV2Token)this;
                return new SecurIdV2Token(this);
            }
        }

        /// <summary>
        /// Gets a <see cref="SecurIdV3Token"/> object representing the same token as the current <see cref="SecurIdToken"/> object.
        /// </summary>
        /// <remarks>
        /// If the current <see cref="SecurIdToken"/> object is an instance of the <see cref="SecurIdV3Token"/> class, then the current instance is returned.  Otherwise, a new instance of the <see cref="SecurIdV3Token"/> class is returned, initialized with the values from the current <see cref="SecurIdToken"/> object.
        /// </remarks>
        public SecurIdV3Token AsV3Token
        {
            get
            {
                if (this is SecurIdV3Token)
                    return (SecurIdV3Token) this;
                return new SecurIdV3Token(this);
            }
        }

        /// <summary>
        /// Gets a <see cref="SecurIdSdtidToken"/> object representing the same token as the current <see cref="SecurIdToken"/> object.
        /// </summary>
        /// <remarks>
        /// If the current <see cref="SecurIdToken"/> object is an instance of the <see cref="SecurIdSdtidToken"/> class, then the current instance is returned.  Otherwise, a new instance of the <see cref="SecurIdSdtidToken"/> class is returned, initialized with the values from the current <see cref="SecurIdToken"/> object.
        /// </remarks>
        public SecurIdSdtidToken AsSdtidToken
        {
            get
            {
                if (this is SecurIdSdtidToken)
                    return (SecurIdSdtidToken) this;
                return new SecurIdSdtidToken(this);
            }
        }

        #endregion

    }
}
