﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ursa.Extensions;

namespace ursa
{
    /// <summary>
    /// Represents a SecurID® token encoded in the Version 2 format
    /// </summary>
    public class SecurIdV2Token : SecurIdToken
    {

        #region Fields

        /// <exclude />
        protected ushort _decryptedSeedHash;
        /// <exclude />
        protected ushort _deviceIdHash;

        /// <exclude />
        protected const int VerChars = 1;
        /// <exclude />
        protected const int TokenBitsPerChar = 3;
        /// <exclude />
        protected const int MinTokenBits = 189;
        /// <exclude />
        protected const int MaxTokenBits = 255;
        /// <exclude />
        protected const int MaxTokenChars = (MaxTokenBits / TokenBitsPerChar);
        /// <exclude />
        protected const int MinTokenChars = ((MinTokenBits / TokenBitsPerChar) + SerialChars + VerChars + ChecksumChars);
        /// <exclude />
        protected const int ChecksumBits = 15;
        /// <exclude />
        protected const int ChecksumChars = (ChecksumBits / TokenBitsPerChar);
        /// <exclude />
        protected const int BinEncBits = 189;
        /// <exclude />
        protected const int BinEncChars = (BinEncBits / TokenBitsPerChar);

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdV2Token"/> class representing an empty token.
        /// </summary>
        public SecurIdV2Token() : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdV2Token"/> class with the value of an existing token.
        /// </summary>
        /// <param name="token">A <see cref="SecurIdToken"/> object representing a SecurID® token.  Ignored if <see langword="null"/>.</param>
        /// <overloads>Initializes a new instance of the <see cref="SecurIdV2Token"/> class.</overloads>
        public SecurIdV2Token(SecurIdToken token) : base(token)
        {
            _decryptedSeedHash = 0;
            _deviceIdHash = 0;
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Decrypts the token using the supplied password and Device ID (Serial Number).
        /// </summary>
        /// <param name="pass">String containing the password for the token. If the token does not require a password, an empty string should be passed.</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token.</param>
        /// <exception cref="ursaException">Thrown if the decryption fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override void Decrypt(string pass, string devid)
        {
            if (IsEmpty)
                throw new InvalidOperationException("Cannot decrypt; no valid token has been loaded.");

            ushort device_id_hash = 0;

            devid = ValidateDeviceIDForDecryption(devid);
            pass = ValidatePasswordForDecryption(pass);

            byte[] key = GetEncryptionKey(pass, devid, out device_id_hash);
            if (key.Length > ursaEncryption.AesBlockSize)
                throw new ursaException(ursaError.DecryptionFailed, "Hash key returned by GetV2HashKey is too long.");

            if (IsDeviceIdRequired && device_id_hash != this._deviceIdHash)
                throw new ursaException(ursaError.DeviceIdInvalid);

            _decryptedSeed = ursaEncryption.AesECBDecrypt(key, EncryptedSeed);
            ushort computed_mac = ursaEncryption.ComputeShortMAC(DecryptedSeed, ursaEncryption.AesKeySize);

            if (computed_mac != _decryptedSeedHash)
                throw new ursaException(ursaError.DecryptionFailed);
        }

        /// <summary>
        /// Creates a string representation of the token, which is encrypted using the supplied password and device ID.
        /// </summary>
        /// <param name="pass">String containing the new password for the token. If an empty string is supplied, the exported token will not require a password</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token, if any.</param>
        /// <returns>A string representation of the token Version 2 format.</returns>
        /// <exception cref="ursaException">Thrown if the export fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override string Export(string pass, string devid)
        {
            if (!HasDecryptedSeed)
                throw new ursaException(ursaError.TokenNeedsDecrypted);

            devid = ScrubDeviceID(devid);

            ushort device_id_hash = 0;

            if (pass.Length > 0)
                _bFlagPasswordPrompt = true;
            if (devid.Length > 0)
                _bFlagSnPrompt = true;

            StringBuilder sbEncoded = new StringBuilder();

            sbEncoded.Append("2");
            if (sbEncoded.Length < VerChars)
                sbEncoded.Append('\0', VerChars - sbEncoded.Length);

            sbEncoded.Append(Serial);
            if (sbEncoded.Length < VerChars + SerialChars)
                sbEncoded.Append('\0', VerChars + SerialChars - sbEncoded.Length);


            byte[] binEncBits = new byte[BinEncChars];
            binEncBits.SetAll(0);


            byte[] key = GetEncryptionKey(pass, devid, out device_id_hash);
            byte[] encryptedSeed = ursaEncryption.AesECBEncrypt(key, DecryptedSeed);

            encryptedSeed.CopyTo(0, binEncBits, 0, Math.Min(ursaEncryption.AesKeySize, encryptedSeed.Length));

            ushort flags = 0;
            flags |= (ushort) (IntervalSeconds == 60 ? 0x03 : 0);  // Bit 0-1 (0x03)
                                                                   // Bit 2 (0x04) is ignored
            flags |= (ushort) (IsLocalPin ? 0x08 : 0);           // Bit 3 (0x08)
            flags |= (ushort) (IsPinRequiredForTokencode ? 0x10 : 0);             // Bit 4 (0x10)
                                                                                  // Bit 5 (0x20) is ignored
            flags |= (ushort) ((((uint) TokencodeLength - 1) << 6) & 0x1C0); // Bit 6-8 (0x1C0)
            flags |= (ushort) (UsesTimeDerivedSeeds ? 0x200 : 0); // Bit 9 (0x200)
            flags |= (ushort) (HasModeFlag ? 0x400 : 0);             // Bit 10 (0x400)
            flags |= (ushort) (UsesAppDerivedSeeds ? 0x800 : 0);  // Bit 11 (0x800)
            flags |= (ushort) (IsDeviceIdRequired ? 0x1000 : 0);        // Bit 12 (0x1000)
            flags |= (ushort) (IsPasswordRequired ? 0x2000 : 0);  // Bit 13 (0x2000)
            flags |= (ushort) (Has128BitKey ? 0x4000 : 0);          // Bit 14 (0x4000)
            binEncBits.SetBits(128, 16, flags);

            binEncBits.SetBits(144, 14, ConvertToSecuridDate(ExpirationDate));
            binEncBits.SetBits(159, 15, ursaEncryption.ComputeShortMAC(DecryptedSeed, ursaEncryption.AesKeySize));
            binEncBits.SetBits(174, 15, device_id_hash);
            sbEncoded.Append(binEncBits.EncodeBitsIntoChars(BinEncBits, TokenBitsPerChar));

            byte[] macBits = new byte[2];
            macBits.SetBits(0, 15, ursaEncryption.ComputeShortMAC(sbEncoded.ToString().ToCharArray().ToByteArray(), sbEncoded.Length));
            sbEncoded.Append(macBits.EncodeBitsIntoChars(ChecksumBits, TokenBitsPerChar));

            return sbEncoded.ToString();
        }

        #endregion


        #region Protected Methods

        /// <summary>
        /// Decodes the Version 2 string representation of a token.  This method is called by <see cref="SecurIdToken.Parse(string, int)"/> to initialize the current instance based on the values in the string representation.
        /// </summary>
        /// <param name="token">String containing a Version 2 representation of a SecurID® token.</param>
        /// <param name="tokenNumber">This value is always ignored, because Version 2 strings cannot contain more than one token.</param>
        /// <exception cref="ursaException">Thrown if the string can not be decoded.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        protected override void DecodeToken(string token, int tokenNumber)
        {
            Regex rgxV2Chars = new Regex("^(?<token>[0-9-]+)", RegexOptions.None);
            Match match = rgxV2Chars.Match(token);
            if (!match.Success)
                throw new ursaException(ursaError.TokenStringInvalid);
            token = match.Groups["token"].Value.Replace("-", "");

            if (token.Length < MinTokenChars || token.Length > MaxTokenChars)
                throw new ursaException(ursaError.TokenStringInvalid, String.Format("Length of token string is invalid.  Expected length {0} to {1} characters; actual length {2} characters.", MinTokenChars, MaxTokenChars, token.Length));

            /* the last 5 digits provide a checksum for the rest of the string */
            byte[] checksumBits = token.DecodeCharsIntoBits(token.Length - ChecksumChars, 15, TokenBitsPerChar);
            ushort iTokenMac = checksumBits.GetBits(0, 15);
            ushort iTokenComputedMac = ursaEncryption.ComputeShortMAC(token.ToCharArray().ToByteArray(), token.Length - ChecksumChars);

            if (iTokenMac != iTokenComputedMac)
                throw new ursaException(ursaError.ChecksumFailed);

            _version = token[0] - '0';
            _sSerial = token.Substring(VerChars, SerialChars);

            _encryptedSeed = new byte[ursaEncryption.AesKeySize];
            _encryptedSeed.SetAll(0);
            byte[] binEncBits = token.DecodeCharsIntoBits(VerChars + SerialChars, BinEncBits, TokenBitsPerChar);
            binEncBits.CopyTo(0, _encryptedSeed, 0, Math.Min(ursaEncryption.AesKeySize, binEncBits.Length));


            ushort flags = binEncBits.GetBits(128, 16);
            _iIntervalSeconds = (flags & 0x03) != 0 ? 60 : 30;  // Bit 0-1 (0x03)
                                                                // Bit 2 (0x04) is ignored
            _bFlagLocalPin = ((flags & 0x08) != 0);             // Bit 3 (0x08)
            _bFlagAddPin = ((flags & 0x10) != 0);               // Bit 4 (0x10)
                                                                // Bit 5 (0x20) is ignored
            _iDigits = ((flags & 0x1C0) >> 6) + 1;              // Bit 6-8 (0x1C0)
            _bFlagTimeDerivedSeeds = ((flags & 0x200) != 0);    // Bit 9 (0x200)
            _bFlagMode = ((flags & 0x400) != 0);                // Bit 10 (0x400)
            _bFlagAppDerivedSeeds = ((flags & 0x800) != 0);     // Bit 11 (0x800)
            _bFlagSnPrompt = ((flags & 0x1000) != 0);           // Bit 12 (0x1000)
            _bFlagPasswordPrompt = ((flags & 0x2000) != 0);     // Bit 13 (0x2000)
            _bFlag128Bit = ((flags & 0x4000) != 0);             // Bit 14 (0x4000)

            _dtExpiration = ConvertSecuridDate(binEncBits.GetBits(144, 14));
            _decryptedSeedHash = binEncBits.GetBits(159, 15);
            _deviceIdHash = binEncBits.GetBits(174, 15);
        }

        /// <summary>
        /// Calculate an encryption key for a given password and device ID (serial number)
        /// </summary>
        /// <param name="pass">String containing the password for the token.</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token.</param>
        /// <param name="deviceIdHash">Upon return, contains the hash value for the device ID.</param>
        /// <returns>An array of bytes containing the key.</returns>
        protected byte[] GetEncryptionKey(string pass, string devid, out ushort deviceIdHash)
        {
            byte[] magic = { 0xd8, 0xf5, 0x32, 0x53, 0x82, 0x89 };
            int devid_len = IsSmartPhone ? 40 : 32;
            if (devid.Length > 0 && devid.Length != devid_len)
                throw new ArgumentException(String.Format("Expecting device ID of length 0 or {0}.", devid_len), "devid");
            byte[] key = new byte[pass.Length + devid.Length + magic.Length + 1];


            int pos = 0;
            key.SetAll(0);
            if (!String.IsNullOrWhiteSpace(pass))
            {
                pass.ToCharArray().ToByteArray().CopyTo(0, key, 0, pass.Length);
                pos += pass.Length;
            }

            int devIdIdx = pos;
            deviceIdHash = 0;
            if (!String.IsNullOrWhiteSpace(devid))
            {
                /*
        		 * For iPhone/Android ctf strings, the device ID takes up
        		 * 40 bytes and consists of hex digits + zero padding.
        		 *
        		 * For other ctf strings (e.g. --blocks), the device ID takes
        		 * up 32 bytes and consists of decimal digits + zero padding.
        		 *
        		 * If this seed isn't locked to a device, we'll just hash
        		 * 40 (or 32) zero bytes, below.
        		 */
                char[] carrDevID = devid.ToCharArray();
                for (int i = 0; i < carrDevID.Length; i++)
                {
                    if ((_version == 1 && Char.IsDigit(carrDevID[i])) ||
                        (_version >= 2 && !Uri.IsHexDigit(carrDevID[i])))
                        continue;
                    key[pos++] = (byte) (Char.ToUpper(carrDevID[i]));
                }

                deviceIdHash = ursaEncryption.ComputeShortMAC(key, devIdIdx, carrDevID.Length);
            }


            magic.CopyTo(0, key, pos, magic.Length);
            pos += magic.Length;

            return ursaEncryption.ComputeMAC(key, pos);
        }

        #endregion


        #region Properties

        /// <summary>
        /// Returns <see langword="true"/> if this <see cref="SecurIdV2Token"/> object does not contain a valid token; <see langword="false"/> otherwise.
        /// </summary>
        public override bool IsEmpty
        {
            get
            {
                return base.IsEmpty || !HasEncryptedSeed;
            }
        }

        #endregion
    }
}
