﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ursa.Extensions;

namespace ursa
{

    /// <summary>
    /// Represents a SecurID® token encoded in the Version 3 format
    /// </summary>
    public class SecurIdV3Token : SecurIdToken
    {

        #region Fields

        /// <exclude />
        public const int V3DevIdChars = 48;
        /// <exclude />
        public const int V3NonceBytes = 16;
        /// <exclude />
        public const int V3PayloadBytes = 0xb0;
        /// <exclude />
        public const int V3TotalBytes = 0x123;

        /// <exclude />
        public const byte V3_ADDPIN_OFF = 0x1f;
        /// <exclude />
        public const byte V3_ADDPIN_ON = 0x21;

        /// <exclude />
        protected byte[] _nonce_devid_hash;
        /// <exclude />
        protected byte[] _nonce_devid_pass_hash;
        /// <exclude />
        protected byte[] _nonce;
        /// <exclude />
        protected byte[] _enc_payload;
        /// <exclude />
        protected byte[] _mac;
        /// <exclude />
        protected byte[] _unk0;
        /// <exclude />
        protected byte[] _res0;
        /// <exclude />
        protected byte[] _res1;
        /// <exclude />
        protected byte[] _res2;

        #endregion


        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdV3Token"/> class representing an empty token.
        /// </summary>
        public SecurIdV3Token() : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SecurIdV3Token"/> class with the value of an existing token.
        /// </summary>
        /// <param name="token">A <see cref="SecurIdToken"/> object representing a SecurID® token.  Ignored if <see langword="null"/>.</param>
        /// <overloads>Initializes a new instance of the <see cref="SecurIdV3Token"/> class.</overloads>
        public SecurIdV3Token(SecurIdToken token) : base(token)
        {
            _nonce_devid_hash = null;
            _nonce_devid_pass_hash = null;
            _nonce = null;
            _enc_payload = null;
            _mac = null;
            _unk0 = null;
            _res0 = null;
            _res1 = null;
            _res2 = null;
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Decrypts the token using the supplied password and Device ID (Serial Number).
        /// </summary>
        /// <param name="pass">String containing the password for the token. If the token does not require a password, an empty string should be passed.</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token.</param>
        /// <exception cref="ursaException">Thrown if the decryption fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override void Decrypt(string pass, string devid)
        {
            if (IsEmpty)
                throw new InvalidOperationException("Cannot decrypt; no valid token has been loaded.");

            devid = ValidateDeviceIDForDecryption(devid);
            pass = ValidatePasswordForDecryption(pass);

            byte[] hash;

            hash = ComputeHash("", devid, _nonce);
            if (hash.CompareTo(_nonce_devid_hash, ursaEncryption.Sha256HashSize) != 0)
                throw new ursaException(ursaError.DeviceIdInvalid);

            hash = ComputeHash(pass, devid, _nonce);
            if (hash.CompareTo(_nonce_devid_pass_hash, ursaEncryption.Sha256HashSize) != 0)
                throw new ursaException(ursaError.DecryptionFailed);

            byte[] hmacKey = DeriveKey(pass, devid, _nonce, 0);
            hash = ursaEncryption.Sha256Hmac(hmacKey, ToBytes(false));
            if (hash.CompareTo(_mac, ursaEncryption.Sha256HashSize) != 0)
                throw new ursaException(ursaError.ChecksumFailed);

            hash = DeriveKey(pass, devid, _nonce, 1);
            byte[] payload = ursaEncryption.AesDecrypt(hash, _nonce, _enc_payload);

            int iPos = 0;

            if (iPos + 16 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _sSerial = new string(payload.ToCharArray(iPos, 16), 0, 16).TrimEnd('\0');
            iPos += 16;

            if (iPos + ursaEncryption.AesKeySize > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _decryptedSeed = new byte[ursaEncryption.AesKeySize];
            payload.CopyTo(iPos, _decryptedSeed, 0, ursaEncryption.AesKeySize);
            iPos += ursaEncryption.AesKeySize;

            if (iPos + 2 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _unk0 = new byte[2];
            payload.CopyTo(iPos, _unk0, 0, 2);
            iPos += 2;

            if (iPos + 1 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _bFlagMode = (payload[iPos++] > 0);

            if (iPos + 1 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _iDigits = (int) (payload[iPos++]);

            if (iPos + 1 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _bFlagAddPin = (payload[iPos++] != V3_ADDPIN_OFF);

            if (iPos + 1 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _iIntervalSeconds = (int) (payload[iPos++]);

            if (iPos + 2 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _res0 = new byte[2];
            payload.CopyTo(iPos, _res0, 0, 2);
            iPos += 2;

            if (iPos + 5 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            byte[] birthDate = new byte[5];
            payload.CopyTo(iPos, birthDate, 0, 5);
            _dtBirth = ConvertV3TokenDate(birthDate);
            iPos += 5;

            if (iPos + 3 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _res1 = new byte[3];
            payload.CopyTo(iPos, _res1, 0, 3);
            iPos += 3;

            if (iPos + 5 > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            byte[] expDate = new byte[5];
            payload.CopyTo(iPos, expDate, 0, 5);
            _dtExpiration = ConvertV3TokenDate(expDate);
            iPos += 5;

            if (iPos + 0x6b > payload.Length)
                throw new ursaException(ursaError.DecryptionFailed);
            _res2 = new byte[0x6b];
            payload.CopyTo(iPos, _res2, 0, 0x6b);
            iPos += 0x6b;

            if (Serial.Length != SecurIdToken.SerialChars)
                throw new ursaException(ursaError.DecryptionFailed);

            _bFlagAppDerivedSeeds = true;
            _bFlag128Bit = true;
        }

        /// <summary>
        /// Creates a string representation of the token, which is encrypted using the supplied password and device ID.
        /// </summary>
        /// <param name="pass">String containing the new password for the token. If an empty string is supplied, the exported token will not require a password</param>
        /// <param name="devid">String containing the Device ID/Serial Number for the token, if any.</param>
        /// <returns>A string representation of the token Version 3 format.</returns>
        /// <exception cref="ursaException">Thrown if the export fails.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        public override string Export(string pass, string devid)
        {
            if (!HasDecryptedSeed)
                throw new ursaException(ursaError.TokenNeedsDecrypted);

            devid = ScrubDeviceID(devid);

            #region Build the payload

            byte[] payload = new byte[V3PayloadBytes];
            payload.SetAll(0);

            int iPos = 0;
            Serial.ToCharArray().ToByteArray().CopyTo(0, payload, iPos, Serial.Length);
            iPos += 16;

            DecryptedSeed.CopyTo(0, payload, iPos, ursaEncryption.AesKeySize);
            iPos += ursaEncryption.AesKeySize;

            payload[iPos++] = 1;
            payload[iPos++] = 1;

            payload[iPos++] = (byte)(HasModeFlag ? 1 : 0);
            payload[iPos++] = (byte)TokencodeLength;
            payload[iPos++] = (IsPinRequiredForTokencode ? V3_ADDPIN_ON : V3_ADDPIN_OFF);
            payload[iPos++] = (byte)IntervalSeconds;

            iPos += 2;

            ConvertToV3TokenDate(BirthDate).CopyTo(0, payload, iPos, 5);
            iPos += 5;

            iPos += 3;

            ConvertToV3TokenDate(ExpirationDate).CopyTo(0, payload, iPos, 5);
            iPos += 5;

            iPos += 0x6b;

            payload.SetAll(0x10, iPos, 0x10);

            #endregion

            #region Build the token

            // Get a nonce
            byte[] nonce = ursaEncryption.GetRandomBytes(V3NonceBytes);

            // Encrypt the payload
            byte[] key = DeriveKey(pass, devid, nonce, 1);
            byte[] payloadEncrypted = ursaEncryption.AesEncrypt(key, nonce, payload);

            // Compute the hashes
            byte[] nonce_devid_hash = ComputeHash("", devid, nonce);
            byte[] nonce_devid_pass_hash = ComputeHash(pass, devid, nonce);

            // Build the token
            byte[] result = new byte[V3TotalBytes];
            result.SetAll(0);
            
            iPos = 0;
            result[iPos++] = (byte)3;
            result[iPos++] = (byte)(pass.Length > 0 ? 1 : 0);
            result[iPos++] = (byte)(devid.Length > 0 ? 1 : 0);

            nonce_devid_hash.CopyTo(0, result, iPos, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            nonce_devid_pass_hash.CopyTo(0, result, iPos, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            nonce.CopyTo(0, result, iPos, V3NonceBytes);
            iPos += V3NonceBytes;

            payloadEncrypted.CopyTo(0, result, iPos, V3PayloadBytes);
            iPos += V3PayloadBytes;

            // Compute the HMAC
            byte[] hmacKey = DeriveKey(pass, devid, nonce, 0);
            byte[] mac = ursaEncryption.Sha256Hmac(hmacKey, result, 0, result.Length - ursaEncryption.Sha256HashSize);

            // Copy the HMAC to the token
            mac.CopyTo(0, result, iPos, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            #endregion

            // Return the token as a Base64 string
            return System.Net.WebUtility.UrlEncode(Convert.ToBase64String(result));
        }

        #endregion


        #region Protected Methods

        /// <summary>
        /// Decodes the Version 3 string representation of a token.  This method is called by <see cref="SecurIdToken.Parse(string, int)"/> to initialize the current instance based on the values in the string representation.
        /// </summary>
        /// <param name="token">String containing a Version 3 representation of a SecurID® token.</param>
        /// <param name="tokenNumber">This value is always ignored, because Version 3 strings cannot contain more than one token.</param>
        /// <exception cref="ursaException">Thrown if the string can not be decoded.  See <see cref="ursaException.ErrorCode" qualifyHint="true"/> for the specific reason for the failure.</exception>
        protected override void DecodeToken(string token, int tokenNumber)
        {
            token = System.Net.WebUtility.UrlDecode(token);

            byte[] tokenBytes = Convert.FromBase64String(token);
            int iPos = 0;

            if (iPos >= tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _version = tokenBytes[iPos++];

            if (iPos >= tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _bFlagPasswordPrompt = (tokenBytes[iPos++] > 0);

            if (iPos >= tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _bFlagSnPrompt = (tokenBytes[iPos++] > 0);

            if (iPos + ursaEncryption.Sha256HashSize > tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _nonce_devid_hash = new byte[ursaEncryption.Sha256HashSize];
            tokenBytes.CopyTo(iPos, _nonce_devid_hash, 0, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            if (iPos + ursaEncryption.Sha256HashSize > tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _nonce_devid_pass_hash = new byte[ursaEncryption.Sha256HashSize];
            tokenBytes.CopyTo(iPos, _nonce_devid_pass_hash, 0, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            if (iPos + V3NonceBytes > tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _nonce = new byte[V3NonceBytes];
            tokenBytes.CopyTo(iPos, _nonce, 0, V3NonceBytes);
            iPos += V3NonceBytes;

            if (iPos + V3PayloadBytes > tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _enc_payload = new byte[V3PayloadBytes];
            tokenBytes.CopyTo(iPos, _enc_payload, 0, V3PayloadBytes);
            iPos += V3PayloadBytes;

            if (iPos + ursaEncryption.Sha256HashSize > tokenBytes.Length)
                throw new ursaException(ursaError.TokenStringInvalid);
            _mac = new byte[ursaEncryption.Sha256HashSize];
            tokenBytes.CopyTo(iPos, _mac, 0, ursaEncryption.Sha256HashSize);
            iPos += ursaEncryption.Sha256HashSize;

            if (Version != 0x03)
                throw new ursaException(ursaError.TokenVersionInvalid);
        }


        /// <summary>
        /// Convert the current <see cref="SecurIdV3Token"/> object into an array of bytes.
        /// </summary>
        /// <param name="includeMac">If <see langword="true"/>, the message authentication code (MAC) will be included in the resulting value; if <see langword="false"/>, the MAC will be excluded.</param>
        /// <returns>An array of bytes representing the current <see cref="SecurIdV3Token"/> object.</returns>
        protected byte[] ToBytes(bool includeMac)
        {
            int iPos = 0;
            byte[] tokenBytes = new byte[3 + _nonce_devid_hash.Length + _nonce_devid_pass_hash.Length + _nonce.Length + _enc_payload.Length + (includeMac ? _mac.Length : 0)];
            tokenBytes[iPos++] = (byte) _version;
            tokenBytes[iPos++] = IsPasswordRequired ? (byte) 1 : (byte) 0;
            tokenBytes[iPos++] = IsDeviceIdRequired ? (byte) 1 : (byte) 0;

            _nonce_devid_hash.CopyTo(0, tokenBytes, iPos, _nonce_devid_hash.Length);
            iPos += _nonce_devid_hash.Length;

            _nonce_devid_pass_hash.CopyTo(0, tokenBytes, iPos, _nonce_devid_pass_hash.Length);
            iPos += _nonce_devid_pass_hash.Length;

            _nonce.CopyTo(0, tokenBytes, iPos, _nonce.Length);
            iPos += _nonce.Length;

            _enc_payload.CopyTo(0, tokenBytes, iPos, _enc_payload.Length);
            iPos += _enc_payload.Length;

            if (includeMac)
            {
                _mac.CopyTo(0, tokenBytes, iPos, _mac.Length);
                iPos += _mac.Length;
            }

            return tokenBytes;
        }

        /// <summary>
        /// Converts a SecurID® Version 3 date value (5 bytes) to a DateTime instance.
        /// </summary>
        /// <param name="v3date">An array of bytes representing a date/time in the SecurID® Version 3 date format. See remarks section for details.</param>
        /// <returns>A DateTime instance initialized to the date/time represented by <paramref name="v3date"/>.</returns>
        /// <remarks>
        /// In Version 3 tokens, the date is a count of "v3 ticks" since 1970-01-01. There are 337500 v3 ticks per day, or 3.90625 v3 ticks per second.  The number of v3 ticks is stored as a 40-bit value (five bytes, big-endian order).
        /// </remarks>
        /// <seealso cref="ConvertToV3TokenDate"/>
        protected DateTime ConvertV3TokenDate(byte[] v3date)
        {
            if (v3date == null || v3date.Length < 5)
                throw new ArgumentException("Expecting an array of 5 bytes", "data");

            long v3Ticks =
                ((long) v3date[0] << 32) |
                ((long) v3date[1] << 24) |
                ((long) v3date[2] << 16) |
                ((long) v3date[3] << 8) |
                ((long) v3date[4] << 0);

            // In v3 tokens, the date is a count of "v3 ticks" since 1970-01-01. 
            // There are 337500 v3 ticks per day, or 3.90625 v3 ticks per second
            return new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddSeconds(v3Ticks / 3.90625);
        }

        /// <summary>
        /// Converts a DateTime instance to a SecurID® Version 3 date value.
        /// </summary>
        /// <param name="time">The DateTime instance to be converted.</param>
        /// <returns>An array of five bytes representing a date/time in the SecurID® Version 3 date format. See remarks section for details.</returns>
        /// <remarks>
        /// In Version 3 tokens, the date is a count of "v3 ticks" since 1970-01-01. There are 337500 v3 ticks per day, or 3.90625 v3 ticks per second.  The number of v3 ticks is stored as a 40-bit value (five bytes, big-endian order).
        /// </remarks>
        /// <seealso cref="ConvertV3TokenDate"/>
        protected byte[] ConvertToV3TokenDate(DateTime time)
        {
            // In v3 tokens, the date is a count of "v3 ticks" since 1970-01-01. 
            // There are 337500 v3 ticks per day, or 3.90625 v3 ticks per second
            long v3Ticks = (long) ((time - (new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc))).TotalSeconds * 3.90625);

            byte[] result = new byte[5];
            result.SetAll(0);

            result[0] = (byte)((v3Ticks >> 32) & 0xFF);
            result[1] = (byte) ((v3Ticks >> 24) & 0xFF);
            result[2] = (byte) ((v3Ticks >> 16) & 0xFF);
            result[3] = (byte) ((v3Ticks >> 8) & 0xFF);
            result[4] = (byte) ((v3Ticks >> 0) & 0xFF);

            return result;
        }

        /// <summary>
        /// Calculate a hash value for a given password, device ID (serial number), and salt value
        /// </summary>
        /// <param name="pass">A string containing a password for the token.</param>
        /// <param name="devid">A string containing a device ID / serial number for the token.</param>
        /// <param name="salt">A string containing a salt value.</param>
        /// <returns>An array of bytes containing the calculated hash.</returns>
        /// <exception cref="ArgumentException">Thrown if <paramref name="salt"/> is null or empty.</exception>
        protected byte[] ComputeHash(string pass, string devid, byte[] salt)
        {
            if (pass == null)
                pass = "";

            if (devid == null)
                devid = "";

            if (salt == null || salt.Length == 0)
                throw new ArgumentException("Salt is null or empty.", "salt");

            byte[] hash_buf = new byte[salt.Length + SecurIdV3Token.V3DevIdChars + pass.Length];

            hash_buf.SetAll(0);

            salt.CopyTo(0, hash_buf, 0, salt.Length);

            if (!String.IsNullOrEmpty(devid))
                devid.ToCharArray().ToByteArray().CopyTo(0, hash_buf, salt.Length, Math.Min(SecurIdV3Token.V3DevIdChars, devid.Length));

            if (!String.IsNullOrEmpty(pass))
                pass.ToCharArray().ToByteArray().CopyTo(0, hash_buf, salt.Length + SecurIdV3Token.V3DevIdChars, pass.Length);

            return ursaEncryption.Sha256Hash(hash_buf);
        }

        /// <summary>
        /// Create an encryption key based on a password, device ID (serial number), salt value, and key number
        /// </summary>
        /// <param name="pass">A string containing a password for the token.</param>
        /// <param name="devid">A string containing a device ID / serial number for the token.</param>
        /// <param name="salt">A string containing a salt value.</param>
        /// <param name="keyId">A number indicating which key to create; must be either 0 or 1.</param>
        /// <returns>An array of bytes containing the new key.</returns>
        /// <exception cref="ArgumentException">Thrown if <paramref name="keyId"/> is a value other than 0 or 1, or if <paramref name="salt"/> is null or empty.</exception>
        protected byte[] DeriveKey(string pass, string devid, byte[] salt, int keyId)
        {
            if (pass == null)
                pass = "";

            if (devid == null)
                devid = "";

            if (keyId < 0 || keyId > 1)
                throw new ArgumentException("Key identifier must be 0 or 1.", "keyId");
            else if (salt == null || salt.Length == 0)
                throw new ArgumentException("Salt is null or empty.", "salt");

            uint i;
            byte[] key0 = { 0xd0, 0x14, 0x43, 0x3c, 0x6d, 0x17, 0x9f, 0xeb, 0xda, 0x09, 0xab, 0xfc, 0x32, 0x49, 0x63, 0x4c };
            byte[] key1 = { 0x3b, 0xaf, 0xff, 0x4d, 0x91, 0x8d, 0x89, 0xb6, 0x81, 0x60, 0xde, 0x44, 0x4e, 0x05, 0xc0, 0xdd };
            byte[] key = (keyId > 0) ? key1 : key0;

            byte[] buf0 = new byte[SecurIdV3Token.V3DevIdChars + key.Length + salt.Length + pass.Length];
            byte[] buf1 = new byte[buf0.Length / 2];

            buf0.SetAll(0);

            if (pass.Length > 0)
                pass.ToCharArray().ToByteArray().CopyTo(0, buf0, 0, pass.Length);

            if (devid.Length > 0)
                devid.ToCharArray().ToByteArray().CopyTo(0, buf0, pass.Length, Math.Min(SecurIdV3Token.V3DevIdChars, devid.Length));

            key.CopyTo(0, buf0, pass.Length + SecurIdV3Token.V3DevIdChars, key.Length);
            salt.CopyTo(0, buf0, pass.Length + SecurIdV3Token.V3DevIdChars + key.Length, salt.Length);


            // PBKDF2 password is "every 2nd byte of the input"
            for (i = 1; i < buf0.Length; i += 2)
                buf1[i / 2] = buf0[i];

            return ursaEncryption.PBKDF2Sha256GetBytes(ursaEncryption.Sha256HashSize, buf1, salt, 1000);
        }

        #endregion


        #region Public Properties

        /// <summary>
        /// Returns <see langword="true"/> if this <see cref="SecurIdV3Token"/> object does not contain a valid token; <see langword="false"/> otherwise.
        /// </summary>
        public override bool IsEmpty
        {
            get
            {
                return _nonce == null || _enc_payload == null;
            }
        }


        #endregion


    }
}
