﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using ursa.Extensions;

namespace ursa
{
    internal class ursaEncryption
    {
        public const int AesBlockSize = 16;
        public const int AesKeySize = 16;

        public const int Sha256BlockSize = 64;
        public const int Sha256HashSize = 32;

        public static int CalculateBase64Size(int dataLen)
        {
            // base64 produces 4 output characters (each carrying 6 bits of data) for
            // up to 3 input bytes(each carrying 8 bits of data). Plus one terminating
            // null at the end.
            return (4 * (dataLen + 2) / 3) + 1;
        }

        public static byte[] GetRandomBytes(int len)
        {
            byte[] result = new byte[len];
            RNGCryptoServiceProvider crypt = new RNGCryptoServiceProvider();
            crypt.GetBytes(result);
            return result;
        }


        #region AES Encryption/Decryption Methods

        public static byte[] AesEncrypt(byte[] key, byte[] iv, byte[] data)
        {
            byte[] encrypted;
            using (AesManaged aes = new AesManaged())
            {
                aes.Key = key;
                aes.IV = iv;
                aes.BlockSize = ursaEncryption.AesBlockSize * 8;
                ICryptoTransform aesEnc = aes.CreateEncryptor();


                using (MemoryStream msEncrypt = new MemoryStream())
                {
                    using (CryptoStream csEncrypt = new CryptoStream(msEncrypt, aesEnc, CryptoStreamMode.Write))
                    {
                        csEncrypt.Write(data, 0, data.Length);
                        encrypted = msEncrypt.ToArray();
                    }
                }
            }
            return encrypted;
        }

        public static byte[] AesDecrypt(byte[] key, byte[] iv, byte[] data)
        {
            byte[] decrypted;
            using (AesManaged aes = new AesManaged())
            {
                aes.Key = key;
                aes.IV = iv;
                aes.BlockSize = ursaEncryption.AesBlockSize * 8;
                ICryptoTransform aesDec = aes.CreateDecryptor();


                using (MemoryStream msDecrypt = new MemoryStream())
                {
                    using (CryptoStream csDecrypt = new CryptoStream(msDecrypt, aesDec, CryptoStreamMode.Write))
                    {
                        csDecrypt.Write(data, 0, data.Length);
                        decrypted = msDecrypt.ToArray();
                    }
                }
            }
            return decrypted;
        }


        public static byte[] AesECBEncrypt(byte[] key, byte[] data)
        {
            byte[] encrypted;
            using (AesManaged aes = new AesManaged())
            {
                aes.Key = key;
                aes.BlockSize = ursaEncryption.AesBlockSize * 8;
                aes.Mode = CipherMode.ECB;
                aes.Padding = PaddingMode.None;
                ICryptoTransform aesEnc = aes.CreateEncryptor();

                using (MemoryStream msEncrypt = new MemoryStream())
                {
                    using (CryptoStream csEncrypt = new CryptoStream(msEncrypt, aesEnc, CryptoStreamMode.Write))
                    {
                        csEncrypt.Write(data, 0, data.Length);
                    }
                    encrypted = msEncrypt.ToArray();
                }
            }
            return encrypted;
        }

        public static byte[] AesECBDecrypt(byte[] key, byte[] data)
        {
            byte[] decrypted;
            using (AesManaged aes = new AesManaged())
            {
                aes.Key = key;
                aes.BlockSize = ursaEncryption.AesBlockSize * 8;
                aes.Mode = CipherMode.ECB;
                aes.Padding = PaddingMode.None;
                ICryptoTransform aesDec = aes.CreateDecryptor();

                using (MemoryStream msDecrypt = new MemoryStream())
                {
                    using (CryptoStream csEncrypt = new CryptoStream(msDecrypt, aesDec, CryptoStreamMode.Write))
                    {
                        csEncrypt.Write(data, 0, data.Length);
                    }
                    decrypted = msDecrypt.ToArray();
                }
            }
            return decrypted;
        }

        public static byte[] AesCbcHash(byte[] key, byte[] iv, byte[] data)
        {
            byte[] plain = new byte[AesBlockSize];
            byte[] enc = new byte[AesBlockSize];
            for (int i = 0; i < data.Length; i += AesBlockSize)
            {
                plain.SetAll(0);
                data.CopyTo(i, plain, 0, Math.Min(AesBlockSize, data.Length - i));

                if (i == 0)
                    enc = AesEncrypt(key, iv, plain);
                else
                    enc = AesEncrypt(key, enc, plain);
            }
            return enc;
        }

        public static void XorBlock(byte[] target, byte[] source)
        {
            XorBlock(target, source, 0);
        }

        public static void XorBlock(byte[] target, byte[] source, int sourceOffset)
        {
            if (source == null || target == null || source.Length - sourceOffset <= 0 || target.Length == 0)
                return;

            for (int i = 0; i < target.Length && i + sourceOffset < source.Length; i++)
                target[i] ^= source[i + sourceOffset];
        }

        public static bool AesBlocksEqual(byte[] x, byte[] y)
        {
            if (x == null || y == null || x.Length != y.Length)
                return false;

            for (int i = 0; i < x.Length && i < ursaEncryption.AesBlockSize; i++)
            {
                if (x[i] != y[i])
                    return false;
            }

            return true;
        }

        public static void AesEcbEncryptThenXor(byte[] key, byte[] data)
        {
            AesEcbEncryptThenXor(key, 0, data);
        }

        public static void AesEcbEncryptThenXor(byte[] key, uint keyOffset, byte[] data)
        {
            if (keyOffset >= key.Length)
                throw new ArgumentOutOfRangeException("keyOffset");

            byte[] key2;
            if (keyOffset > 0 || key.Length > AesKeySize)
            {
                key2 = new byte[AesKeySize];
                key2.SetAll(0);
                key.CopyTo((int) keyOffset, key2, 0, (int) Math.Min(AesKeySize, key.Length - keyOffset));
            }
            else
                key2 = key;

            byte[] enc = AesECBEncrypt(key2, data);
            XorBlock(data, enc);
        }

        #endregion


        #region SHA1 Encryption Methods

        public static byte[] Sha1Hash(byte[] data)
        {
            using (SHA1Managed sha1 = new SHA1Managed())
            {
                return sha1.ComputeHash(data);
            }

        }

        public static byte[] RsaSha1SignData(RSAParameters key, byte[] data)
        {
            RSACryptoServiceProvider RSA = new RSACryptoServiceProvider();
            RSA.ImportParameters(key);
            return RSA.SignData(data, new SHA1CryptoServiceProvider());
        }

        #endregion


        #region SHA256 Encryption Methods

        public static byte[] Sha256Hash(byte[] data)
        {
            using (SHA256Managed sha256 = new SHA256Managed())
            {
                return sha256.ComputeHash(data);
            }

        }

        public static byte[] Sha256Hmac(byte[] password, byte[] data)
        {
            return Sha256Hmac(password, data, 0, data.Length);
        }

        public static byte[] Sha256Hmac(byte[] password, byte[] data, int offset, int count)
        {
            using (var hmac = new HMACSHA256(password))
            {
                return hmac.ComputeHash(data, offset, count);
            }
        }

        // NOTE: The iteration count should
        // be as high as possible without causing
        // unreasonable delay.  Note also that the password
        // and salt are byte arrays, not strings.  After use,
        // the password and salt should be cleared (with Array.Clear)
        public static byte[] PBKDF2Sha256GetBytes(int dklen, byte[] password, byte[] salt, int iterationCount)
        {
            using (var hmac = new HMACSHA256(password))
            {
                int hashLength = hmac.HashSize / 8;
                if ((hmac.HashSize & 7) != 0)
                    hashLength++;
                int keyLength = dklen / hashLength;
                if ((long) dklen > (0xFFFFFFFFL * hashLength) || dklen < 0)
                    throw new ArgumentOutOfRangeException("dklen");
                if (dklen % hashLength != 0)
                    keyLength++;
                byte[] extendedkey = new byte[salt.Length + 4];
                Buffer.BlockCopy(salt, 0, extendedkey, 0, salt.Length);
                using (var ms = new System.IO.MemoryStream())
                {
                    for (int i = 0; i < keyLength; i++)
                    {
                        extendedkey[salt.Length] = (byte) (((i + 1) >> 24) & 0xFF);
                        extendedkey[salt.Length + 1] = (byte) (((i + 1) >> 16) & 0xFF);
                        extendedkey[salt.Length + 2] = (byte) (((i + 1) >> 8) & 0xFF);
                        extendedkey[salt.Length + 3] = (byte) (((i + 1)) & 0xFF);
                        byte[] u = hmac.ComputeHash(extendedkey);
                        Array.Clear(extendedkey, salt.Length, 4);
                        byte[] f = u;
                        for (int j = 1; j < iterationCount; j++)
                        {
                            u = hmac.ComputeHash(u);
                            for (int k = 0; k < f.Length; k++)
                            {
                                f[k] ^= u[k];
                            }
                        }
                        ms.Write(f, 0, f.Length);
                        Array.Clear(u, 0, u.Length);
                        Array.Clear(f, 0, f.Length);
                    }
                    byte[] dk = new byte[dklen];
                    ms.Position = 0;
                    ms.Read(dk, 0, dklen);
                    ms.Position = 0;
                    for (long i = 0; i < ms.Length; i++)
                    {
                        ms.WriteByte(0);
                    }
                    Array.Clear(extendedkey, 0, extendedkey.Length);
                    return dk;
                }
            }
        }
        #endregion


        #region Message Authentication Code (MAC) Methods

        public static byte[] ComputeMAC(byte[] data, int in_len)
        {
            return ComputeMAC(data, 0, in_len);
        }

        public static byte[] ComputeMAC(byte[] data, int offset, int in_len)
        {

            if (offset + in_len > data.Length)
                throw new ArgumentOutOfRangeException("in_len");


            byte[] work = new byte[AesKeySize];
            work.SetAll(0xff);


            /* padding */
            byte[] pad = new byte[AesKeySize];
            pad.SetAll(0);
            for (int i = in_len * 8, padIdx = pad.Length - 1; i > 0 && padIdx >= 0; i >>= 8, padIdx--)
                pad[padIdx] = (byte) i;

            /* handle the bulk of the input data here */
            bool odd = false;
            int iDataIdx;
            for (iDataIdx = 0; in_len - iDataIdx > AesKeySize; iDataIdx += AesKeySize, odd = !odd)
                AesEcbEncryptThenXor(data, (uint) iDataIdx + (uint) offset, work);

            /* final 0-16 bytes of input data */
            byte[] lastblk = new byte[AesKeySize];
            lastblk.SetAll(0);
            data.CopyTo(iDataIdx + offset, lastblk, 0, in_len - iDataIdx);
            AesEcbEncryptThenXor(lastblk, work);

            /* hash an extra block of zeros, for certain input lengths */
            if (odd)
            {
                byte[] zero = new byte[AesKeySize];
                zero.SetAll(0);
                AesEcbEncryptThenXor(zero, work);
            }

            /* always hash the padding */
            AesEcbEncryptThenXor(pad, work);

            /* run hash over current hash value, then return */
            byte[] result = new byte[AesKeySize];
            work.CopyTo(0, result, 0, work.Length);
            AesEcbEncryptThenXor(work, result);

            return result;
        }

        public static ushort ComputeShortMAC(byte[] data, int in_len)
        {
            return ComputeShortMAC(data, 0, in_len);
        }

        public static ushort ComputeShortMAC(byte[] data, int offset, int in_len)
        {
            byte[] result = new byte[AesBlockSize];

            result = ComputeMAC(data, offset, in_len);
            return (ushort) ((result[0] << 7) | (result[1] >> 1));
        }

        #endregion

    }
}
