﻿/*
 * This file is part of ursa, a tokencode generator compatible with 
 * RSA SecurID® 128-bit (AES) tokens. RSA SecurID® is a registered 
 * trademark held by RSA Security LLC.  Neither the ursa project 
 * nor Samoyed Software, LLC are affiliated with or endorsed by 
 * RSA® (RSA Security LLC). 
 * 
 * Copyright 2017 Samoyed Software, LLC <http://www.samoyedsoftware.com>
 *
 * Portions of ursa are based on 
 * stoken 0.92, Copyright 2012 Kevin Cernekee <http://stoken.sf.net/>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ursa
{
    /// <summary>
    /// The exception that is throw when an operation by attempted by the ursa library fails.
    /// </summary>
    public class ursaException : System.Exception
    {
        /// <exclude />
        private readonly ursaError _errCode;

        /// <summary>
        /// Initializes a new instance of the <see cref="ursaException"/> class, with an <see cref="ErrorCode"/> of <see cref="ursaError.None" qualifyHint="true"/>.
        /// </summary>
        public ursaException() : this(ursaError.None)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ursaException"/> class with a <see cref="ursaError"/> value.
        /// </summary>
        /// <param name="errorCode">A <see cref="ursaError"/> value indicating the type of error.</param>
        public ursaException(ursaError errorCode) : this(errorCode, "", null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ursaException"/> class with a <see cref="ursaError"/> value and error message.
        /// </summary>
        /// <param name="errorCode">A <see cref="ursaError"/> value indicating the type of error.</param>
        /// <param name="message">A string containing the error message</param>
        public ursaException(ursaError errorCode, string message) : this(errorCode, message, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ursaException"/> class with a <see cref="ursaError"/> value, error message, and inner exception.
        /// </summary>
        /// <param name="errorCode">A <see cref="ursaError"/> value indicating the type of error.</param>
        /// <param name="message">A string containing the error message.</param>
        /// <param name="innerException">A <see cref="Exception"/> object representing the inner exception.</param>
        /// <overloads>
        /// Initializes a new instance of the <see cref="ursaException"/> class.
        /// </overloads>
        public ursaException(ursaError errorCode, string message, Exception innerException) : base(message.Length > 0 ? message : errorCode.ToString(), innerException)
        {
            _errCode = errorCode;
        }

        /// <summary>
        /// Gets the <see cref="ursaError"/> value indicating the type of error represented by this instance.
        /// </summary>
        public ursaError ErrorCode
        {
            get
            {
                return _errCode;
            }
        }

        /// <summary>
        /// Gets a human-readable string representation of the <see cref="ErrorCode"/>.
        /// </summary>
        public string ErrorCodeAsString
        {
            get
            {
                return ErrorCodeToString(ErrorCode);
            }
        }

        /// <summary>
        /// Gets a human-readable string representation of a <see cref="ursaError"/> value.
        /// </summary>
        /// <param name="error">A <see cref="ursaError"/> value.</param>
        /// <returns>A string describing the meaning of the value of <paramref name="error"/>.</returns>
        public static string ErrorCodeToString(ursaError error)
        {
            string msg = "";
            switch (error)
            {
                case ursaError.None:
                    msg = "Success.";
                    break;

                case ursaError.GeneralError:
                    msg = "General error.";
                    break;

                case ursaError.XmlInvalid:
                    msg = "XML is invalid.";
                    break;

                case ursaError.TokenStringInvalid:
                    msg = "Token string is invalid.";
                    break;

                case ursaError.TokenVersionInvalid:
                    msg = "Token version is invalid.";
                    break;

                case ursaError.ChecksumFailed:
                    msg = "Checksum validation failed.";
                    break;

                case ursaError.PasswordRequired:
                    msg = "Password is required.";
                    break;

                case ursaError.PinRequired:
                    msg = "PIN is required.";
                    break;

                case ursaError.TokenNeedsDecrypted:
                    msg = "Token must be decrypted by calling the Decrypt method before the requested action can be completed.";
                    break;

                case ursaError.DecryptionFailed:
                    msg = "Decryption failed.";
                    break;

                case ursaError.DeviceIdInvalid:
                    msg = "Device ID is not valid.";
                    break;

                case ursaError.DeviceIdRequired:
                    msg = "Device ID is required.";
                    break;

                case ursaError.FileReadFailed:
                    msg = "File could not be read.";
                    break;

                case ursaError.SdtidMultipleTokens:
                    msg = "SDTID token has multiple TKN elements.";
                    break;

                default:
                    msg = "Unknown error";
                    break;
            }

            return msg;

        }
    }
}
